/**

    Copyright 2007 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.graph.parser;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dom4j.Attribute;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.XPath;
import org.dom4j.io.SAXReader;
import org.dom4j.xpath.DefaultXPath;

public class SvgParser {
	long _processId;
	long _processInstanceId;
	//_svgBean can contains processSvg or processInstanceSvg
	Document _svgDocument = null;
	IProcessInfo _processInfo = null;
	
	private Map<String, String> namespaceMap = null;
	
	private static final Log log = LogFactory.getLog(SvgParser.class);
	
	public static final String NODE_LABEL = "node";
	public static final String EDGE_LABEL = "edge";

	public static final String FONT_SIZE = "font-size".intern();
	public static final String FONT_SIZE_VALUE = "6.00".intern();
	
	public SvgParser(long processId, String svg, IProcessInfo processInfo) throws Exception {
		_processId = processId;
		_processInfo = processInfo;
		initInputSvgBean(svg);	
	}

	public SvgParser(long processId, long processInstanceId, String svg, IProcessInfo processInfo) throws Exception {
		_processId = processId;
		_processInstanceId = processInstanceId;
		_processInfo = processInfo;
		initInputSvgBean(svg);
	}
	
	private void initInputSvgBean(String roughSvg) throws Exception {
		SAXReader reader = new SAXReader();
		InputStream is = new ByteArrayInputStream(roughSvg.getBytes());
		_svgDocument = reader.read(is);
		namespaceMap = new HashMap<String, String>();
		namespaceMap.put("svg", "http://www.w3.org/2000/svg");
		namespaceMap.put("xlink", "http://www.w3.org/1999/xlink");
	}
	
	private Element evalXPathSingleNode( Element doc, String xPathExpressionString){
		XPath xPathExpression = new DefaultXPath(xPathExpressionString);
		xPathExpression.setNamespaceURIs(namespaceMap);
		return (Element)xPathExpression.selectSingleNode(doc);
	}
	
	private List<Element> evalXPathNodes( Element doc, String xPathExpressionString){
		XPath xPathExpression = new DefaultXPath(xPathExpressionString);
		xPathExpression.setNamespaceURIs(namespaceMap);
		return (List<Element>)xPathExpression.selectNodes(doc);
	}
	
	 public List<Element> getFilteredSourceBeanAttributeAsList(String xpath, String attributeName, String attributeValue) {
		    	List<Element> nodes = new ArrayList<Element>();
		    	Element root = (Element)_svgDocument.selectSingleNode("/svg");
		    	List<Element> elements = evalXPathNodes(root, xpath);
		    	if(elements.size() == 0)
		    		return nodes;
		    	
		    	for(Element element : elements){
		    		if(element.attribute(attributeName).getValue().compareTo(attributeValue) == 0)
		    			nodes.add(element);
		    	}
		    	return nodes;
		    }
	
	public String createProcessSvgString() throws Exception {
		try {
			Element svg = (Element)_svgDocument.selectSingleNode("/svg");
			String width = svg.valueOf("@width");
			((Element)svg).attribute("width").setValue(getBiggerWidth(width));			
				
			List<Element> elementList = getFilteredSourceBeanAttributeAsList("//svg:g/svg:g", "class", NODE_LABEL);
			
			String stepName = null;	
			for (Element element : elementList) {
				Element a = evalXPathSingleNode(element, "svg:a");
				a.attribute("href").setValue("javascript:void(0)");			
				stepName = evalXPathSingleNode(element, "svg:title").getText(); 
				Element text = evalXPathSingleNode(element, "svg:a/svg:text");
				// Compatibility check
				if (stepName == null || stepName == "") {
					stepName = text.getText();
				}
				Attribute y = text.attribute("y");
				modifyCoordinateNextText(y.getValue());
				StepInfo step = _processInfo.getStepInfo(stepName);
				String originalName = step.getOriginalStepName();
				element.attribute("id").setValue(stepName);
				
				// Remove process name from step name
				String decoratedStepName = decorateStepName(originalName);
				
				int dotPos = decoratedStepName.indexOf('.');
				if (dotPos != -1) {
					decoratedStepName = decoratedStepName.substring(dotPos + 1);
				}
				
				
				
				text.setText(decoratedStepName);
				Attribute style = text.attribute("style");
				if (style != null) {
					style.setValue("font-size:6.00pt;");
				} else if (text.attribute(FONT_SIZE) != null) {
					text.attribute(FONT_SIZE).setValue(FONT_SIZE_VALUE);
				} else {
					text.addAttribute(FONT_SIZE, FONT_SIZE_VALUE);
				}
					
				text.addAttribute("title", "Step");
				text.addAttribute("onclick", "showDetail(evt, " + step.getStepID() + ");");
				
				if(step.isStartStep() && _processInstanceId == 0) {
					if(step.getStepState() == 1)
						changeNodeState(IProcessInfo.TRANSITION_FAULTED, element);
				}
				
				//add element text linked to subprocess
				SubProcessInfo subProcess = _processInfo.getSubProcessInfo(stepName);
				if (subProcess != null) {
					Element newText = a.addElement("text");
					newText.addAttribute("style", text.attributeValue("style"));
					newText.addAttribute("text-anchor", text.attributeValue("text-anchor"));
					newText.addAttribute("x", text.attributeValue("x"));
					Double subProcessCoordinate = Double.parseDouble(text.attributeValue("y")) + 10;
					newText.addAttribute("y", subProcessCoordinate.toString());	
					newText.addAttribute("onclick", "showSubProcess(evt, " + subProcess.getId() + ", " + step.getStepID() + ");");
					newText.setText(subProcess.getName());
					changeStyle("font-style:", "italic", newText);
					newText.addAttribute("title", "SubProcess");
				}
			}
			
			Element script = _svgDocument.getRootElement().addElement("script");
			setCDATABean(script);			
		} catch(Exception sbe) {		
			log.error("SvgParser::createProcessSvgString: ", sbe);
			throw sbe;
		}
		return _svgDocument.asXML();
	}
	               
	private String decorateStepName(String stepName) {
		
		// Remove process name from step name
		int dotPos = stepName.indexOf('.');
		if (dotPos != -1) {
			stepName = stepName.substring(dotPos + 1);
		}
		
		// Remove also the process version
		int versionPos = stepName.indexOf("_v_");
		if (versionPos != -1) {
			stepName = stepName.substring(0, versionPos);
		}
		
		return stepName;
	}
	
	private void modifyCoordinateNextText(String value) throws Exception {
		List<Element> textList = getFilteredSourceBeanAttributeAsList("//svg:g/svg:g/svg:a/svg:text", "y", value);
		try {
			if(textList.size()>1) {
				double startY = Double.parseDouble(value);
				double distance = 6;
				int counter = 0;
				for(Element elem: textList){
					switch (counter) {
					case 0:		
						counter++;
						elem.attribute("y").setValue(Double.toString(startY));
						continue;
					case 1:
						counter++;
						elem.attribute("y").setValue(Double.toString(startY- distance));
						continue;
					case 2:
						counter = 0;
						elem.attribute("y").setValue(Double.toString(startY+ distance));
						continue;
					}
				}
			}
		} catch(Exception sbe) {		
			log.error("SvgParser::modifyCoordinateNextText: ", sbe);
			throw sbe;
		}
	}
	
	private void setCDATABean(Element script) throws Exception{
		try {
			script.addAttribute("type", "text/javascript");
			String functionshowDetailString = "function showDetail(evt, stepId){\n"
								  				+ "	top.showDetailStep(stepId);\n"
												+"}\n";
			
			String functionshowSubProcessString = "function showSubProcess(evt, processId, stepId){\n";
			if(_processInstanceId == 0)
				functionshowSubProcessString += "	top.showDetailSubProcess('',processId);\n";
			else
				functionshowSubProcessString += "	top.showListSubProcess(stepId);\n";
			functionshowSubProcessString += "}\n";
			script.setText("\n"+ functionshowDetailString +"\n"+ functionshowSubProcessString);
		} catch(Exception sbe) {		
			log.error("SvgParser::getCDATABean: ", sbe);
			throw sbe;
		}
		return;
	}
	
	private String getBiggerWidth(String width) {
		if(width.indexOf("pt") != -1) {
			String widthTemp = width.substring(0, width.indexOf("pt")); 
			int newWidth = Integer.parseInt(widthTemp) + 70; 
			return Integer.toString(newWidth) + "pt";
		} else {
			String widthTemp = width.substring(0, width.indexOf("in")); 
			double newWidth = Double.parseDouble(widthTemp) + 2; 
			return Double.toString(newWidth) + "in";
		}
	}

	public String createProcessInstanceSvgString() throws Exception {
		createProcessSvgString();
		List<Element> elementList = getFilteredSourceBeanAttributeAsList("//svg:g/svg:g", "class", NODE_LABEL);
		String stepName;
		int stateInstanceService;
		for (Element element : elementList) {
			stepName = evalXPathSingleNode(element, "svg:title").getText(); 
			Element text = evalXPathSingleNode(element, "svg:a/svg:text");
			//Compatibility check for old graphs
			if (stepName == null || stepName == "") {
				stepName = text.getText();
			}

			// Retrieve process state
			stateInstanceService = _processInfo.getExecutionStateInstanceService(stepName, _processInstanceId);
			changeNodeState(stateInstanceService, element);
		}		
		return _svgDocument.asXML();
	}
	
	private void changeNodeState(int state, Element element) throws Exception{
		String newColor = null;
		switch (state) {
		case IProcessInfo.TRANSITION_DONE:
			// Green
			newColor = "#99FF99";// "#33FF99"			
			break;
		case IProcessInfo.TRANSITION_ACTIVE:
			// Yellow
			newColor = "#FFFF99";
			break;
		case IProcessInfo.TRANSITION_FAULTED:
			newColor = "red";
			break;
		case IProcessInfo.TRANSITION_WARNING:
			newColor = "#FFA500"; //#FF6600 #FF6103 #FFA500
			break;
		}
		
		if (newColor == null) {
			return;
		}
		
		Element shape = evalXPathSingleNode(element, "svg:a/svg:polygon");

		if (shape != null) {
			List<Element> nodeList = evalXPathNodes(element,  "svg:a/svg:path");
			
			if (nodeList.isEmpty()) {
				// Mantain border color
				changeFillColor(newColor, shape);
			} else {
				// Color also the border
				changeFillColor(newColor, shape);
				changeStrokeColor(newColor, shape);

				for (Element polyLine : nodeList)
					changeFillColor(newColor, polyLine);
			}
		} else {
			shape =  evalXPathSingleNode(element, "svg:a/svg:ellipse");
			if (shape != null) {
				changeFillColor(newColor, shape);
			}
		}
	}

	private void changeFillColor(String newColor, Element node) throws Exception {
		changeStyle("fill:", newColor, node);
	}
	
	private void changeStrokeColor(String newColor, Element node) throws Exception {
		changeStyle("stroke:", newColor, node);
	}

	private void changeStyle(String attribute, String newStyle, Element node) throws Exception {
		Attribute styleAttr = node.attribute("style");
		String style = (styleAttr != null) ? styleAttr.getValue() : null;
		
		if (style != null){
			int attributePos = style.indexOf(attribute);
			if (attributePos != -1) {
				int attributeEndPos = style.indexOf(';', attributePos);
				style = style.substring(0, attributePos) + attribute + newStyle + style.substring(attributeEndPos);
			}
			else {
				style = style + attribute + newStyle + ";";
			}
		
			try {
				node.attribute("style").setValue(style);
			} catch (Exception e) {
				log.error("SvgParser::changeColor: ", e);
				throw e;
			}
		}else{
			// NEW SVG STYLE fill and stroke are two separate attribute
			if (attribute.startsWith("fill")){
				node.attribute("fill").setValue(newStyle);
			}else if (attribute.startsWith("stroke")){
				node.attribute("stroke").setValue(newStyle);
			}
		}
	}
	
}
