package org.spagic3.attachments.fs;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.attachments.IAttachmentStore;
import org.spagic3.constants.SpagicConstants;



public class FileSystemAttachmentStore implements IAttachmentStore {

	private static final int BUFFER_SIZE = 4096;
	
	private String attachmentStoreId = null;

	private static String nodeId = System.getProperty(SpagicConstants.SPAGIC_NODE_IDENTIFIER); 
	
	private Logger log = LoggerFactory.getLogger(FileSystemAttachmentStore.class);
	public String rootFolder = null;
	
	
	
	
	public String getAttachmentStoreId() {
		return attachmentStoreId;
	}

	public void setAttachmentStoreId(String attachmentStoreId) {
		this.attachmentStoreId = attachmentStoreId;
	}
	
	public FileSystemAttachmentStore(){
		rootFolder = System.getProperty(SpagicConstants.ATTACHMENT_STORE_FYLESYSTEM_ROOT);
		
		if (rootFolder == null){
			// set configured root folder
			rootFolder = System.getProperty(SpagicConstants.SPAGIC_HOME_PROPERTY) 
									+ "/store/attachments/";
		}

		// create directory if doesn't exist
		File f = new File(rootFolder);
		if (!f.exists()){
			boolean ok = f.mkdirs();
			//TODO: manage directory cration fault
			if (!ok){
				log.error("Unable to create rootFolder[" + rootFolder + "]");
			}
		}
			
		// check if readonly
		if (!f.canWrite()){
			//TODO: manage directory writeable fault
			log.error("Unable to write into rootFolder[" + rootFolder + "]");
		}

		//modify backslash into slash
		rootFolder = rootFolder.replace ('\\', '/');
		
		// add slash if doesn't exist
		if (!rootFolder.endsWith("/")){ 
			rootFolder += "/";
		}
	}
		
	@Override
	public void delete(String id) {

		// get file
		File aFile = getAttachmentFile(id);

		// if file doas'nt exist --> return
		if (aFile == null) 
			return;

		boolean ok = aFile.delete();
		// check if successful
		if (!ok){
			log.error("Cannot delete ID ["+id+": file must be deletable -->" + aFile.getAbsolutePath());
		}
		return;		
	}

	@Override
	public DataHandler load(String id) {
		
		// get file
		File aFile = getAttachmentFile(id);
		
		// if file doas'nt exist --> return
		if (aFile == null) 
			return null;
		
		// create inputStream and return it as DataHandler
		FileDataSource fds = new FileDataSource(aFile);
		
		DataHandler dh = new DataHandler(fds);
		
		return dh;
	}

	@Override
	public String store(String name, String correlationID, DataHandler value) {
		
		// check name, correlationID, value
		if (name==null || "".equals(name.trim())){
			log.error("Attachment name is null or empty");
			return null;
		}

		try{
			if (value==null || value.getDataSource() == null || value.getDataSource().getInputStream()==null){
				log.error("Attachment data Handler is null or invalid");
				return null;
			}
		} catch (Exception e){
			log.error("Attachment data Handler is null or invalid", e);
			return null;
		}
		
		
		// create unique file name
		String relativeFileName = getUniqueFileName(name, correlationID);

		// create absolute filename
		String absoluteFilename = rootFolder; 
		if (!rootFolder.endsWith("/")) absoluteFilename += "/";
		absoluteFilename += relativeFileName;
		
		// create subdir if doesn't exist
		File aFile = new File (absoluteFilename);
		File aDir = aFile.getParentFile();
		if (!aDir.exists()) {
			boolean ok = aDir.mkdirs();
			if (!ok){
				log.error("Unable to create rootFolder[" + rootFolder + "]");
				return null;
			}
		}
		
		//create smartPointer
		String smartPointer = "[FS]/" + relativeFileName;
		
		//copy input stream into file
		try {
			if ( copyFile(value.getDataSource().getInputStream(), aFile)){
				return smartPointer;
			}
		} catch (IOException e) {
			log.error("Failed to create file[" + absoluteFilename + "]",e);
		}
		
		return null;
	}

// **********************************************************************	
// ************************** helper methods ****************************
// **********************************************************************	

	private boolean copyFile(InputStream from, File to) {
		FileOutputStream fos = null;
		try {
			fos = new FileOutputStream(to);
			byte[] buffer = new byte[BUFFER_SIZE];
			int byte_count = 0;
			while ((byte_count = from.read(buffer)) > 0) {
				fos.write(buffer, 0, byte_count);
			}
			return true;
		} catch (IOException ex) {
			try {
				if( fos != null)
					fos.close();
			} catch (Exception ex1) {}
			try {
				if (to.exists())
					to.delete();
			} catch (Exception ex1) {}
			return false;
		} finally {
			try {
				fos.close();
			} catch (Exception ex) {}
			try {
				from.close();
			} catch (Exception ex) { }
		}
	}	
	
    private String getUniqueFileName(String name, String correlationID){

    	StringBuffer sb = new StringBuffer();
    	
    	// convention for creating file name:  nodeId/YYYYMMDD/HH24/corelationID_mmssSSS_randomvalue_name
    	
    	// nodeID
    	sb.append(nodeId).append("/");
    	
    	// YYYYMMDD/HH24  
    	SimpleDateFormat formatter = new SimpleDateFormat ("yyyyMMdd/HH");
    	Date aDate = new java.util.Date();
    	sb.append(formatter.format(aDate)).append("/");

    	//corelationID_mmssSSS_randomvalue_name
		if (correlationID!=null && !("".equals(correlationID.trim()))){
			sb.append(correlationID).append("_");
		}
		formatter = new SimpleDateFormat ("mmssSSS");
		sb.append(formatter.format(aDate)).append("_");
		sb.append(""+(int)(Math.random()*Integer.MAX_VALUE)).append("_");
		sb.append(name);
		
		return sb.toString();
    }

    private File getAttachmentFile(String id){
		// check id
		if (id == null || "".equals(id.trim())){
			log.error("Attachment ID is null or empty");
			return null;
		}
		
		//id must start with "[FS]"
		if (!id.startsWith("[FS]")){
			log.error("Attachment ID invalid: must start with [FS] -->" + id);
			return null;
		}

		// get absolute file name from smartPointer: rootFolder + id without prefix "[FS]/" 
		String relativeFilename = id.substring("[FS]/".length());
		String absololuteFilename = rootFolder + relativeFilename;

		// check if exists file inside repository
		File aFile = new File(absololuteFilename);
		if (!aFile.exists()){
			log.error("Attachment ID invalid ["+id+": file must exists -->" + absololuteFilename);
			return null;
		}
		
    	return aFile; 
		
    }

    
    public static void main (String [] args){
    	try{
    	FileSystemAttachmentStore fsas = new FileSystemAttachmentStore();
    	
    	FileDataSource fds = new FileDataSource("c:\\post.html");
    	DataHandler dh = new DataHandler(fds);
    	
    	String id = fsas.store("post.html", "correlationID0000001", dh);
    	
    	DataHandler dh2 = fsas.load(id);
    	InputStream is = dh2.getDataSource().getInputStream();
    	fsas.copyFile(is, new File("c:\\temp\\post.html"));
    	
    	fsas.delete(id);
    	
    	} catch (Exception e){
    		e.printStackTrace();
    	}
    	
    }
}
