package org.spagic3.client.api;

import java.io.File;

/**
 * Interface <code>IAttachment</code> is the contract that must be
 * fulfilled by a class modeling an attachment meant to be added to
 * the ClientMessage. The attachment is supposed to be set to contain
 * either a File or byte array, <b>but not both</b>.
 *
 * @author <a href="mailto:saint@eng.it">Gian Uberto Lauri</a>
 * @version $Revision$
 * @see ClientMessage
 */
public interface IAttachment {


	/**
	 * <code>setData</code> Sets the data of the attachment as a File
	 *
	 * @param data a <code>File</code> the File containing the data
	 * @param mimeType a <code>String</code> the mime type of the data
	 * @exception <code>Error</code> when the progremmer 
	 * attempts to set again the data on the same IAttachment. 
	 */
	public void setData(File data, String mimeType);

	/**
	 * <code>setData</code> Sets the data of the attachment as a byte array
	 *
	 * @param data a <code>byte</code> the byte array containing the data
	 * @param mimeType a <code>String</code> the mime type of the data
	 * @exception <code>Error</code> when the progremmer 
	 * attempts to set again the data on the same IAttachment. 
	 */
	public void setData(byte data[], String mimeType);

	/**
	 * <code>isDataAFile</code> predicate true when the IAttachment
	 * was built using a File (the setData(File) was used).
	 *
	 * @return a <code>boolean</code> true if the data was set with
	 * the setData(File), false otherwise.
	 */
	public boolean isDataAFile();
	
	/**
	 * <code>getDataAsByteArray</code> Returns the data as a byte
	 * array. Valid only if the IAttachment was build using a byte
	 * array (the setData(byte[]) was used).
	 *
	 * @return a <code>byte[]</code> The byte array used to set the data in
	 * the IAttachment
	 * @exception <code>Error</code> if the IAttachment was built using a File (the
	 * setData(File) was used).
	 */
	public byte[] getDataAsByteArray() throws Error;
	
	/**
	 * <code>getDataAsFile</code> Returns the data as a File. Valid only if
	 * the IAttachment was build using a File (the setData(File) was used).
	 *
	 * @return a <code>File</code> The File used to set the data in the
	 * IAttachment
	 * @exception <code>Error</code> if the IAttachment was built using a byte array
	 * (the setData(byte[]) was used).
	 */
	public File getDataAsFile() throws Error;

	/**
	 * <code>getMimeType</code> returns the mime type of the
	 * data. Implementors should take care to return a valid mime
	 * type.
	 *
	 * @return a <code>String</code> containing the mime type
	 */
	public String getMimeType();
}
