package org.spagic3.client.api;

import java.io.File;

/**
 * Class <code>MessageAttachment</code> a concrete implementation of
 * the IAttachment interface, models an attachment supplied either as
 * a File or byte array.
 *
 * @author <a href="mailto:saint@eng.it">Gian Uberto Lauri</a>
 * @version $Revision$
 */
public class MessageAttachment implements IAttachment {

	private File file;
	private byte[] byteArray;
	private String mimeType;

	/**
	 * <code>getMimeType</code> returns the mime type of the data. If
	 * no 'valid' mime type was supplied, the methor returns
	 * 'application/octet-stream'. Vaild is a mime type that's not
	 * null and different from a string of blanks. 
	 *
	 * @return a <code>String</code> the mime type of the data.
	 * @exception <code>Error</code> when the progremmer 
	 * attempts to use this method before stetting 
	 * the data.
	 */
	public String getMimeType() {
		if (file == null && byteArray == null )
			throw 
				new Error(
					"No data was set for this attachment, no mime type available");

		return (mimeType == null || mimeType.trim().equals("")) 
			? "application/octet-stream"
			: mimeType;
	}

	/**
	 * <code>setData</code> Sets the data of the attachment as a File. 
	 *
	 * @param data a <code>File</code> the File containing the data
	 * @param mimeType a <code>String</code> the mime type of the data
	 * @exception <code>Error</code> when the progremmer 
	 * attempts to set again the data on the same IAttachment. 
	 */
	public void setData(File data, String mimeType) {
		if( byteArray != null || file != null)
			throw new Error("Yon can not set again the data for this attachment!");

		file = data;
		this.mimeType = mimeType;

	}

	/**
	 * <code>setData</code> Sets the data of the attachment as a byte
	 * array.
	 *
	 * @param data a <code>byte</code> the byte array containing the
	 * data
	 * @param mimeType a <code>String</code> the mime type of the data
	 * @exception <code>Error</code> when the progremmer 
	 * attempts to set again the data on the same IAttachment. 
	 */
	public void setData(byte[] data, String mimeType) {
		if( byteArray != null || file != null)
			throw new Error("Yon can not set again the data for this attachment!");

		byteArray = data;
		this.mimeType = mimeType;
	}

	/**
	 * <code>getDataAsByteArray</code> Returns the data as a byte
	 * array. Valid only if the IAttachment was build using a byte
	 * array (the setData(byte[]) was used).
	 *
	 * @return a <code>byte[]</code> The byte array used to set the
	 * data in the IAttachment
	 * @exception <code>Error</code> if the IAttachment was built using a File (the
	 * setData(File) was used).
	 */
	public byte[] getDataAsByteArray() throws Error {
		if (byteArray == null)
			throw new Error("This attachment was not built with a byte array!");
		return byteArray;
	}

	/**
	 * <code>getDataAsFile</code> Returns the data as a File. Valid
	 * only if the IAttachment was build using a File (the
	 * setData(File) was used).
	 *
	 * @return a <code>File</code> The File used to set the data in
	 * the IAttachment
	 * @exception <code>Error</code> if the IAttachment was built using a byte
	 * array (the setData(byte[]) was used).
	 */
	public File getDataAsFile() throws Error {
		if (file == null)
			throw new Error("This attachment was not built with a File!");
		return file;
	}

	/**
	 * <code>isDataAFile</code> predicate true when the IAttachment
	 * was built using a File (the setData(File) was used).
	 *
	 * @return a <code>boolean</code> true if the data was set with
	 * the setData(File), false otherwise.
	 */
	public boolean isDataAFile() {
		
		return file!=null && byteArray == null;
	}


}
