package org.spagic3.components.zip;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.core.BaseSpagicService;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;

public class UnzipComponent extends BaseSpagicService {

	public static final String PROPERTY_COPY_ATTACHMENTS = "copyAttachments";
	public static final String PROPERTY_ATTACHMENTS_REGEX = "attachmentsRegEx";

	protected final int BUFFER = 2048;
	protected Logger logger = LoggerFactory.getLogger(UnzipComponent.class);

	/**
	 * Regular expressions for filtering attachment names
	 */
	protected String attachmentsRegEx = null;

	public void init() {
		try {
			setCopyAttachments(propertyConfigurator
					.getBoolean(PROPERTY_COPY_ATTACHMENTS));
			attachmentsRegEx = propertyConfigurator.getString(
					PROPERTY_ATTACHMENTS_REGEX, null);
		} catch (Exception e) {
			logger.error(e.getMessage(), e);
			throw new RuntimeException("Cannot instantiate Unzip Component", e);
		}
	}

	@Override
	public boolean run(Exchange exchange, Message in, Message out)
			throws Exception {

		Pattern regexPattern = null;
		if (attachmentsRegEx != null) {
			regexPattern = Pattern.compile(attachmentsRegEx);
		}

		try {
			StringBuffer sb = new StringBuffer();
			Map<String, Object> attachments = in.getAttachments();
			int unzippedFiles = 0;
			
			sb.append("<ATTACHMENTS>");
			
			for (String attachmentKey : attachments.keySet()) {
				DataHandler dh = (DataHandler) attachments.get(attachmentKey);
				logger.debug("Datahandler name: " + dh.getName());
				logger.debug("Datasource name: " + dh.getDataSource().getName());				
				
				boolean matches = true;
				// Optionally check compliance with the file filter
				if (regexPattern != null) {
					Matcher m = regexPattern.matcher(dh.getName());
					matches = m.matches();
				}
				
				if (matches) {
					ZipInputStream zis = new ZipInputStream(dh.getInputStream());

					BufferedOutputStream dest = null;
					ZipEntry entry;
					while ((entry = zis.getNextEntry()) != null) {
						logger.debug("Extracting: " + entry.getName());
						int count;
						byte data[] = new byte[BUFFER];
						
						// write the files to the disk
						File tempFile = File.createTempFile("spagic-", null);
						tempFile.deleteOnExit();
						FileOutputStream fos = new FileOutputStream(tempFile);
						dest = new BufferedOutputStream(fos, BUFFER);
						while ((count = zis.read(data, 0, BUFFER)) != -1) {
							dest.write(data, 0, count);
						}
						dest.flush();
						dest.close();
						
						FileDataSource fds = new FileDataSource(tempFile);
						DataHandler newdh = new DataHandler(fds);
						out.addAttachment(entry.getName(), newdh);
						
						sb.append("<ATTACHMENT>");
						sb.append(entry.getName());
						sb.append("</ATTACHMENT>");
						unzippedFiles++;
						
						// The file cannot be deleted because it's referenced by the datasource
						//tempFile.delete();					
					}
					zis.close();
				}			
			}
			sb.append("</ATTACHMENTS>");
			
			if (unzippedFiles != 0) {
				out.setBody(sb.toString());
			} else {
				fail(exchange, new RuntimeException("No files unzipped"));
				return false;
			}
			
		} catch (Throwable th) {
			logger.error("Error unzipping files", th);
			fail(exchange, new RuntimeException(th.getMessage()));
//			out.setBody("<ERROR>" + th.getMessage() + "</ERROR>");
		}
		return true;	
	}

}
