package org.spagic3.connectors.filesystem.writer;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.activation.DataHandler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.filesystem.adapters.IFileAdapter;
import org.spagic3.core.AbstractSpagicConnector;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;


public class FileSystemWriterConnector extends AbstractSpagicConnector {

	public static final String PROPERTY_DIRECTORY = "directory";
	public static final String PROPERTY_OUTPUT = "output";
	
	public static final String OUTPUT_BODY = "body";
	public static final String OUTPUT_ATTACHMENTS = "attachments";
		
    private static final String filePrefix = "spagic-";
    private static final String fileSuffix = ".xml";

    protected Logger logger = LoggerFactory.getLogger(FileSystemWriterConnector.class);
    
    private boolean autoCreateDirectory = true;
	private File directory;
    private String output;
    
    private ConcurrentHashMap<String, IFileAdapter> marshalers = new ConcurrentHashMap<String, IFileAdapter>();
    private String marshalerId;
    
	public void init() {
		String directoryString = propertyConfigurator.getString(PROPERTY_DIRECTORY);
		this.directory = new File(directoryString);
		this.output = propertyConfigurator.getString(PROPERTY_OUTPUT, OUTPUT_BODY);
		this.marshalerId = propertyConfigurator.getString("marshaller", "DEFAULT");
	}

	public void start() throws Exception {
		super.start();
		if (directory == null) {
			throw new RuntimeException("You must specify the directory");
		}
		if (autoCreateDirectory) {
			directory.mkdirs();
		}
		if (!directory.isDirectory()) {
			throw new RuntimeException(
					"The directory property must be a directory but was: "
							+ directory);
		}
	}
    
	public void stop() throws Exception {
		super.stop();
	}
    
    public File getDirectory() {
        return directory;
    }

  
    public void setDirectory(File directory) {
        this.directory = directory;
    }	


	@Override
	public boolean run(Exchange exchange, Message inMessage, Message outMessage)
			throws Exception {
		BufferedWriter bw = null;
        File newFile = null;
        boolean success = false;
        try {
        	// Check if we have to write a file containing the message body
        	// or if we have to write out the attachments
        	if (output.equalsIgnoreCase(OUTPUT_BODY)) {
        		IFileAdapter adapter = marshalers.get(marshalerId);
                String name = adapter.getOutputFileName(exchange, inMessage);
                if (name == null) {
                    newFile = File.createTempFile(filePrefix, fileSuffix, directory);
                } else {
                    newFile = new File(directory, name);
                }
                if (!newFile.getParentFile().exists() && autoCreateDirectory) {
                    newFile.getParentFile().mkdirs();
                }
                if (logger.isDebugEnabled()) {
                    logger.debug("Writing to file: " + newFile.getCanonicalPath());
                }
                bw = new BufferedWriter(new FileWriter(newFile, false));
                adapter.writeMessage(exchange, inMessage, bw);
        	} else {
        		// Iterate over all attachments and save them
        		Map<String, Object> attachments = inMessage.getAttachments();
        		for (String key : attachments.keySet()) {
        			DataHandler dh = (DataHandler)attachments.get(key);
        			
//					File f = new File(directory, dh.getName());
					File f = new File(directory, key);
					InputStream inputStream = dh.getInputStream();
					OutputStream out = new FileOutputStream(f);
					byte buf[] = new byte[1024];
					int len;
					while ((len = inputStream.read(buf)) > 0)
						out.write(buf, 0, len);
					out.close();
					inputStream.close();
        		}
        	}
        	
            success = true;
        }catch (IOException e) {
			throw new IllegalStateException(e.getMessage(), e);
        }catch (Exception genericException) {
        	throw new IllegalStateException(genericException.getMessage(), genericException);

		} finally {
            if (bw != null) {
                try {
                	bw.close();
                } catch (IOException e) {
                    logger.error("Caught exception while closing stream on error: " + e, e);
                }
            }
            //cleaning up incomplete files after things went wrong
            if (!success && (newFile != null)) {
            	try {
            		logger.debug("An error occurred while writing file " + newFile.getCanonicalPath() + ", deleting the invalid file");
            		if (!newFile.delete())
            			logger.warn("Unable to delete the file " + newFile.getCanonicalPath() + " after an error had occurred");
            	} catch (IOException e) {
	                    logger.error("Caught exception while closing stream on error: " + e, e);
	            }
            }
		}	
		return success;
	}
    	
	public void setMarshaler (IFileAdapter adapter) {
		this.marshalers.put(adapter.getAdapterId(), adapter);
	}	
	public void unsetMarshaler(IFileAdapter marshaler) {
		this.marshalers.remove(marshaler.getAdapterId());
	}

}
