/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.ftp;

import java.io.FileFilter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Date;
import java.util.Timer;
import java.util.concurrent.ConcurrentHashMap;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.ftp.marshaler.FileMarshaler;
import org.spagic3.core.AbstractSpagicConnector;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;

/**
 * A polling endpoint which looks for a file or files in a directory
 * and sends the files into the JBI bus as messages, deleting the files
 * by default when they are processed.
 *
 */
public class FtpPollerComponent extends AbstractSpagicConnector implements IFtpReader {
	
	protected Logger log = LoggerFactory.getLogger(FtpPollerComponent.class);
	
	public static final String PROPERTY_PASSIVEMODE = "passiveMode";
	public static final String PROPERTY_PERIOD = "period";
    
    // Container for polling parameters
	private FtpParameters ftpParams;
	private Timer timerService;
	private PollingTask task;

    private FileFilter filter;
    private ConcurrentHashMap<String, FileMarshaler> marshalers = new ConcurrentHashMap<String, FileMarshaler>();
    private String marshalerId;
    
	public void init() {
		ftpParams = new FtpParameters();
		FTPClientPool clientPool = null;
			
		marshalerId = propertyConfigurator.getString(PROPERTY_MARSHALER, null);
		
		try{
			clientPool = createClientPool();
			clientPool.setBinaryMode(propertyConfigurator.getBoolean(PROPERTY_BINARYMODE, true));
			clientPool.setUsername(propertyConfigurator.getString(PROPERTY_USERNAME, "anonymous"));
			clientPool.setPassword(propertyConfigurator.getString(PROPERTY_PASSWORD, null));
			if(propertyConfigurator.getBoolean(PROPERTY_PASSIVEMODE, null)!=null)
				clientPool.setPassiveMode(propertyConfigurator.getBoolean(PROPERTY_PASSIVEMODE));
			clientPool.setHost(propertyConfigurator.getString(PROPERTY_HOST));
			clientPool.setPort(propertyConfigurator.getInteger(PROPERTY_PORT, 21));		
			setClientPool(clientPool);
		}catch (Exception e) {
			log.error("Error creating FTPClientPool", e);
		}
			
		setDeleteFile(propertyConfigurator.getBoolean(PROPERTY_DELETEFILE, true));
		setRecursive(propertyConfigurator.getBoolean(PROPERTY_RECURSIVE, true));
		setChangeWorkingDirectory(propertyConfigurator.getBoolean(PROPERTY_CHANGEWORKINGDIRECTORY));
		try{
			URI uri = createUri(clientPool, propertyConfigurator.getString(PROPERTY_FOLDER,""));
			setUri(uri);
		}catch (URISyntaxException e) {
			log.error("Error creating URI" , e);
		}
		setStateless(propertyConfigurator.getBoolean(PROPERTY_STATELESS, true));
		setPeriod(propertyConfigurator.getLong(PROPERTY_PERIOD, new Long(10000)));
	}
	
	public void setMarshaler(FileMarshaler marshaler) {
		this.marshalers.put(marshaler.getMarshalerId(), marshaler);
	}

	public void unsetMarshaler(FileMarshaler marshaler) {
		this.marshalers.remove(marshaler.getMarshalerId());
	}

	private URI createUri(FTPClientPool clientPool, String folder) throws URISyntaxException{
		if (clientPool.getUsername().equalsIgnoreCase("anonymous")){
			clientPool.setPassword("anonymous@anonymous.com");
		}
		String ftpUri = "ftp://"+clientPool.getUsername()+(clientPool.getPassword()!=null ? ":"+clientPool.getPassword() : "")+"@"+clientPool.getHost()+":"+clientPool.getPort()+"/"+folder;
		URI uri = new URI(ftpUri);
		return uri;
	}

	@Override
    public void start() throws Exception {
		super.start();
		// Activate polling thread
		timerService = new Timer();
		
        task = new PollingTask(this, target);
        timerService.schedule(task, new Date(), getPeriod());
    }
    
    @Override
	public void stop() throws Exception {
    	super.stop();
    	// Component deactivated
		timerService.cancel();
		timerService = null;
		
	}

    // Properties
    //-------------------------------------------------------------------------
    /**
     * @return the clientPool
     */
    public FTPClientPool getClientPool() {
        return ftpParams.clientPool;
    }

    /**
     * @param clientPool the clientPool to set
     */
    public void setClientPool(FTPClientPool clientPool) {
    	ftpParams.clientPool = clientPool;
    }
    
    public long getPeriod() {
		return ftpParams.period;
	}
	
	public void setPeriod(long period) {
		ftpParams.period = period;
    }

    /**
     * @return the uri
     */
    public URI getUri() {
        return ftpParams.uri;
    }

    /**
     * @param uri the uri to set
     */
    public void setUri(URI uri) {
    	ftpParams.uri = uri;
    }
    
    public String getWorkingPath() {
        return (ftpParams.uri != null && ftpParams.uri.getPath() != null) ? ftpParams.uri.getPath() : ".";
    }

    public FileFilter getFilter() {
        return filter;
    }

    /**
     * Sets the optional filter to choose which files to process
     */
    public void setFilter(FileFilter filter) {
        this.filter = filter;
    }

    /**
     * Returns whether or not we should delete the file when its processed
     */
    public boolean isDeleteFile() {
        return ftpParams.deleteFile;
    }

    public void setDeleteFile(boolean deleteFile) {
    	ftpParams.deleteFile = deleteFile;
    }

    public boolean isRecursive() {
        return ftpParams.recursive;
    }

    public void setRecursive(boolean recursive) {
    	ftpParams.recursive = recursive;
    }

    public FileMarshaler getMarshaler() {
    	if (!marshalers.containsKey(DEFAULT_MARSHALER)) {
			log.error("Default Marshaler not found");
			return null;
    	}
    	
    	if (marshalerId == null) {
			log.info("Using Default Marshaler");
    		return marshalers.get(DEFAULT_MARSHALER);
    	} else if (!marshalers.containsKey(marshalerId)) {
			log.info("Marshaler with id \"" + marshalerId + "\" not found. Using Default Marshaler");
    		return marshalers.get(DEFAULT_MARSHALER);
    	} else {
    		return marshalers.get(marshalerId);
    	}    	
    }

    public void setChangeWorkingDirectory(boolean changeWorkingDirectory) {
    	ftpParams.changeWorkingDirectory = changeWorkingDirectory;
    }
    
    public boolean isChangeWorkingDirectory() {
    	return ftpParams.changeWorkingDirectory;
    }

    public boolean isStateless() {
        return ftpParams.stateless;
    }

    public void setStateless(boolean stateless) {
    	ftpParams.stateless = stateless;
    }

    // Implementation methods
    //-------------------------------------------------------------------------

    public String getLocationURI() {
        return getUri().toString();
    }

    @Override
    public void process(Exchange exchange) {
    	try{
    		task.process(exchange);
    	}catch (Exception e) {
			log.error("Error ", e);
		}
    }

    protected FTPClientPool createClientPool() throws Exception {
        FTPClientPool pool = new FTPClientPool();
        pool.afterPropertiesSet();
        return pool;
    }

	@Override
	public Exchange getFtpExchange() {
		return createInOnlyExchange();
	}
	
	@Override
	public Message getMessage(Exchange exchange) {
		Message message = exchange.getIn(true);
		return message;
	}
	
	/**
	 * Send a message to the NMR
	 * 
	 * @param messageStringBuffer
	 * @throws Exception
	 */
	public void sendFtpExchange(Exchange exchange){
		log.debug("Sending message ");
		send(exchange);
	}

}
