/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.ftp;

import java.io.File;
import java.io.FileFilter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.concurrent.ConcurrentHashMap;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.ftp.marshaler.FileMarshaler;
import org.spagic3.core.BaseSpagicService;
import org.spagic3.core.SpagicUtils;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;

/**
 * A polling endpoint which looks for a file or files in a directory
 * and sends the files into the JBI bus as messages, deleting the files
 * by default when they are processed.
 *
 */
public class FtpReaderComponent extends BaseSpagicService implements IFtpReader {
	
	protected Logger log = LoggerFactory.getLogger(FtpReaderComponent.class);
	
	public static final String PROPERTY_USER_XPATH = "userXPath";
	public static final String PROPERTY_PASSWORD_XPATH = "passwordXPath";
	public static final String PROPERTY_HOST_XPATH = "hostXPath";
	public static final String PROPERTY_PORT_XPATH = "portXPath";
	public static final String PROPERTY_FILE_FILTER_XPATH = "fileFilterXPath";

    private static ThreadLocal<Exchange> exchanges = new ThreadLocal<Exchange>();
    private static ThreadLocal<Message> outMessages = new ThreadLocal<Message>();

    // Container for polling parameters
	private FtpParameters ftpParams;
    private ConcurrentHashMap<String, FileMarshaler> marshalers = new ConcurrentHashMap<String, FileMarshaler>();
    private String marshalerId;

    private FileFilter filter;
    
    private String userXPath;
    private String passwordXPath;
    private String hostXPath;
    private String portXPath;
    private String fileFilterXPath;
    
	public void init() {
		ftpParams = new FtpParameters();
		FTPClientPool clientPool = null;
			
		marshalerId = propertyConfigurator.getString(PROPERTY_MARSHALER, null);
		
		try{
			clientPool = createClientPool();
			clientPool.setBinaryMode(propertyConfigurator.getBoolean(PROPERTY_BINARYMODE, true));
			clientPool.setUsername(propertyConfigurator.getString(PROPERTY_USERNAME, null));
			clientPool.setPassword(propertyConfigurator.getString(PROPERTY_PASSWORD, null));
			if(propertyConfigurator.getBoolean("passiveMode", null)!=null)
				clientPool.setPassiveMode(propertyConfigurator.getBoolean("passiveMode"));
			clientPool.setHost(propertyConfigurator.getString(PROPERTY_HOST));
			clientPool.setPort(propertyConfigurator.getInteger(PROPERTY_PORT, 21));		
			setClientPool(clientPool);
			
			userXPath = propertyConfigurator.getString(PROPERTY_USER_XPATH, null);
		    passwordXPath = propertyConfigurator.getString(PROPERTY_PASSWORD_XPATH, null);
		    hostXPath = propertyConfigurator.getString(PROPERTY_HOST_XPATH, null);
		    portXPath = propertyConfigurator.getString(PROPERTY_PORT_XPATH, null);
		   
		    fileFilterXPath = propertyConfigurator.getString(PROPERTY_FILE_FILTER_XPATH, null);
		}catch (Exception e) {
			log.error("Error creating FTPClientPool", e);
		}
			
		setDeleteFile(propertyConfigurator.getBoolean(PROPERTY_DELETEFILE, true));
		setRecursive(propertyConfigurator.getBoolean(PROPERTY_RECURSIVE, true));
		setChangeWorkingDirectory(propertyConfigurator.getBoolean(PROPERTY_CHANGEWORKINGDIRECTORY));
		try{
			URI uri = createUri(clientPool, propertyConfigurator.getString(PROPERTY_FOLDER,""));
			setUri(uri);
		}catch (URISyntaxException e) {
			log.error("Error creating URI" , e);
		}		
		setStateless(propertyConfigurator.getBoolean(PROPERTY_STATELESS, true));
	}
	
	public void setMarshaler(FileMarshaler marshaler) {
		this.marshalers.put(marshaler.getMarshalerId(), marshaler);
	}

	public void unsetMarshaler(FileMarshaler marshaler) {
		this.marshalers.remove(marshaler.getMarshalerId());
	}

	private URI createUri(FTPClientPool clientPool, String folder) throws URISyntaxException{
		if (clientPool.getUsername().equalsIgnoreCase("anonymous")){
			clientPool.setPassword("anonymous@anonymous.com");
		}
		String ftpUri = "ftp://"+clientPool.getUsername()+(clientPool.getPassword()!=null ? ":"+clientPool.getPassword() : "")+"@"+clientPool.getHost()+":"+clientPool.getPort()+"/"+folder;
		URI uri = new URI(ftpUri);
		return uri;
	}

    // Properties
    //-------------------------------------------------------------------------
    /**
     * @return the clientPool
     */
    public FTPClientPool getClientPool() {
        return ftpParams.clientPool;
    }

    /**
     * @param clientPool the clientPool to set
     */
    public void setClientPool(FTPClientPool clientPool) {
    	ftpParams.clientPool = clientPool;
    }
    
    /**
     * @return the uri
     */
    public URI getUri() {
        return ftpParams.uri;
    }

    /**
     * @param uri the uri to set
     */
    public void setUri(URI uri) {
    	ftpParams.uri = uri;
    }
    
    public String getWorkingPath() {
        return (ftpParams.uri != null && ftpParams.uri.getPath() != null) ? ftpParams.uri.getPath() : ".";
    }

    public FileFilter getFilter() {
        return filter;
    }

    /**
     * Sets the optional filter to choose which files to process
     */
    public void setFilter(FileFilter filter) {
        this.filter = filter;
    }

    /**
     * Returns whether or not we should delete the file when its processed
     */
    public boolean isDeleteFile() {
        return ftpParams.deleteFile;
    }

    public void setDeleteFile(boolean deleteFile) {
    	ftpParams.deleteFile = deleteFile;
    }

    public boolean isRecursive() {
        return ftpParams.recursive;
    }

    public void setRecursive(boolean recursive) {
    	ftpParams.recursive = recursive;
    }

    public FileMarshaler getMarshaler() {
    	if (!marshalers.containsKey(DEFAULT_MARSHALER)) {
			log.error("Default Marshaler not found");
			return null;
    	}
    	
    	if (marshalerId == null) {
			log.info("Using Default Marshaler");
    		return marshalers.get(DEFAULT_MARSHALER);
    	} else if (!marshalers.containsKey(marshalerId)) {
			log.info("Marshaler with id \"" + marshalerId + "\" not found. Using Default Marshaler");
    		return marshalers.get(DEFAULT_MARSHALER);
    	} else {
    		return marshalers.get(marshalerId);
    	}    	
    }

    public void setChangeWorkingDirectory(boolean changeWorkingDirectory) {
    	ftpParams.changeWorkingDirectory = changeWorkingDirectory;
    }
    
    public boolean isChangeWorkingDirectory() {
    	return ftpParams.changeWorkingDirectory;
    }

    public boolean isStateless() {
        return ftpParams.stateless;
    }

    public void setStateless(boolean stateless) {
    	ftpParams.stateless = stateless;
    }

    // Implementation methods
    //-------------------------------------------------------------------------

    public String getLocationURI() {
        return getUri().toString();
    }

    protected FTPClientPool createClientPool() throws Exception {
        FTPClientPool pool = new FTPClientPool();
        pool.afterPropertiesSet();
        return pool;
    }

   
	public Exchange getExchange(){
		return createInOnlyExchange();
	}
	
	/**
	 * Send a message to the NMR
	 * 
	 * @param messageStringBuffer
	 * @throws Exception
	 */
	public void sendFtpExchange(Exchange exchange){
		// Do nothing because the response will be sent at the end of the run method
		log.debug("Sending message ");
	}
	
	public Exchange getFtpExchange() {
		Exchange result = exchanges.get();
		if (result != null) {
			exchanges.remove();
		} else {
			log.error("Unexpected call to getFtpExchange");
		}
		return result;
	}

	public Message getMessage(Exchange exchange) {
		Message result = outMessages.get();
		if (result != null) {
			outMessages.remove();
		} else {
			log.error("Unexpected call to getMessage");
		}
		return result;
	}

	@Override
	public boolean run(Exchange exchange, Message in, Message out)
			throws Exception {
		String fileFilter = null;        
		
        // Retrieve connection parameters dynamically, if required
		if (userXPath != null) {
        	getClientPool().setUsername(SpagicUtils.xpathEvaluateString((String) in.getBody(), userXPath));
		}
		if (passwordXPath != null) {
        	getClientPool().setPassword(SpagicUtils.xpathEvaluateString((String) in.getBody(), passwordXPath));
		}
		if (hostXPath != null) {
        	getClientPool().setHost(SpagicUtils.xpathEvaluateString((String) in.getBody(), hostXPath));
		}
		if (portXPath != null) {
			String port = SpagicUtils.xpathEvaluateString((String) in.getBody(), portXPath);
			if (port != null) {
				try {
					int portInt = Integer.valueOf(port);
		        	getClientPool().setPort(portInt);
				} catch (NumberFormatException nfe) {
					log.error("Error retrieving port from XPath expression: " + portXPath, nfe);
				}
			}
		}
		if (fileFilterXPath != null) {
			fileFilter = SpagicUtils.xpathEvaluateString((String) in.getBody(), fileFilterXPath);
        	setFilter(new EqualityFilter(fileFilter));
		}		
		
		// Store exchange for later use in PollingTask
		exchanges.set(exchange);
		outMessages.set(out);
		
		PollingTask pt = new PollingTask(this, target);
		pt.run();
		if (!pt.isMessageSent()) {
			fail(exchange, new RuntimeException("<ERROR>File " + ((fileFilter != null) ? fileFilter : "") + " not found</ERROR>"));
			return false;
		}
		return true;
	}
	
	public Exchange createExchange() {
		return createInOnlyExchange();
	}
	
	  public class EqualityFilter implements FileFilter {
		
		private String fileName;

		public EqualityFilter(String fileName) {
			super();
			// Remove // pattern at the begin of the filename
			if (fileName.startsWith("\\\\")) {
				fileName = "\\" + fileName.substring(2);
			}
			this.fileName = fileName;
		}

		@Override
		public boolean accept(File pathname) {
			String path = pathname.getPath();
			
			path=path.replace('/', '\\');
			fileName=fileName.replace('/', '\\');

			if (path.startsWith("\\\\")) {
				path = "\\" + path.substring(2);
			}
			
			log.debug("path:" + path + ",fileName:" + fileName);
			return fileName.equals(path) || fileName.contains(path) || path.contains(fileName);
		}
		
	}	

}
