/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.ftp;

import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.net.ftp.FTPClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.ftp.marshaler.FileMarshaler;
import org.spagic3.core.AbstractSpagicConnector;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;

/**
 * An FTP endpoint
 *
 * @version $Revision: $
 * @org.apache.xbean.XBean element="sender"
 */
public class FtpSenderComponent extends AbstractSpagicConnector {
	
	protected Logger log = LoggerFactory.getLogger(FtpSenderComponent.class);
	
	public static final String DEFAULT_MARSHALER = "DEFAULT";

	public static final String PROPERTY_MARSHALER = "marshaler";
	public static final String PROPERTY_UNIQUEFILENAME = "uniqueFileName";
	public static final String PROPERTY_OVERWRITE = "overwrite";
	public static final String PROPERTY_URI = "uri";
	public static final String PROPERTY_UPLOADSUFFIX = "uploadSuffix";
	public static final String PROPERTY_CHECKDUPLICATES = "checkDuplicates";
	public static final String PROPERTY_FOLDER = "folder";

    private FTPClientPool clientPool;
    private String uniqueFileName;
    private boolean overwrite;
    private URI uri;
    private String uploadSuffix;
    private boolean checkDuplicates = true;
    private ConcurrentHashMap<String, FileMarshaler> marshalers = new ConcurrentHashMap<String, FileMarshaler>();
    private String marshalerId;
    
    public void init() {
    	
		marshalerId = propertyConfigurator.getString(PROPERTY_MARSHALER, null);
		
		try{
			clientPool = createClientPool();
			clientPool.setBinaryMode(propertyConfigurator.getBoolean("binaryMode",true));
			clientPool.setUsername(propertyConfigurator.getString("username", "anonymous"));
			clientPool.setPassword(propertyConfigurator.getString("password", null));
			if(propertyConfigurator.getBoolean("passiveMode", null)!=null)
				clientPool.setPassiveMode(propertyConfigurator.getBoolean("passiveMode"));
			clientPool.setHost(propertyConfigurator.getString("host"));
			clientPool.setPort(propertyConfigurator.getInteger("port", 21));		
			setClientPool(clientPool);
		}catch (Exception e) {
			log.error("Error creating FTPClientPool", e);
		}
		
		setUniqueFileName(propertyConfigurator.getString(PROPERTY_UNIQUEFILENAME, "Spagic_"));
		setOverwrite(propertyConfigurator.getBoolean(PROPERTY_OVERWRITE, false));
		
		try{
			URI uri = createUri(clientPool, propertyConfigurator.getString(PROPERTY_FOLDER,""));
			setUri(uri);
		}catch (URISyntaxException e) {
			log.error("Error creating URI" , e);
		}	
		
		
		setUploadSuffix(propertyConfigurator.getString(PROPERTY_UPLOADSUFFIX, ".xml"));
		setCheckDuplicates(propertyConfigurator.getBoolean(PROPERTY_CHECKDUPLICATES, true));
    }
    
	public void setMarshaler(FileMarshaler marshaler) {
		this.marshalers.put(marshaler.getMarshalerId(), marshaler);
	}

	public void unsetMarshaler(FileMarshaler marshaler) {
		this.marshalers.remove(marshaler.getMarshalerId());
	}

    public FileMarshaler getMarshaler() {
    	if (!marshalers.containsKey(DEFAULT_MARSHALER)) {
			log.error("Default Marshaler not found");
			return null;
    	}
    	
    	if (marshalerId == null) {
			log.info("Using Default Marshaler");
    		return marshalers.get(DEFAULT_MARSHALER);
    	} else if (!marshalers.containsKey(marshalerId)) {
			log.info("Marshaler with id \"" + marshalerId + "\" not found. Using Default Marshaler");
    		return marshalers.get(DEFAULT_MARSHALER);
    	} else {
    		return marshalers.get(marshalerId);
    	}    	
    }

    private URI createUri(FTPClientPool clientPool, String folder) throws URISyntaxException{
		if (clientPool.getUsername().equalsIgnoreCase("anonymous")){
			clientPool.setPassword("anonymous@anonymous.com");
		}
		String ftpUri = "ftp://"+clientPool.getUsername()+(clientPool.getPassword()!=null ? ":"+clientPool.getPassword() : "")+"@"+clientPool.getHost()+":"+clientPool.getPort()+"/"+folder;
		URI uri = new URI(ftpUri);
		return uri;
	}

    /**
     * Configures the endpoint from a URI
     */
    public void setUri(URI uri) {
        this.uri = uri;
    }

    public boolean isCheckDuplicates() {
        return checkDuplicates;
    }

    public void setCheckDuplicates(boolean checkDuplicates) {
        this.checkDuplicates = checkDuplicates;
    }
    
   
    // Properties
    //-------------------------------------------------------------------------
    public FTPClientPool getClientPool() {
        return clientPool;
    }

    public void setClientPool(FTPClientPool clientPool) {
        this.clientPool = clientPool;
    }

    public String getUniqueFileName() {
        return uniqueFileName;
    }

    /**
     * Sets the name used to make a unique name if no file name is available on the message.
     *
     * @param uniqueFileName the new value of the unique name to use for generating unique names
     */
    public void setUniqueFileName(String uniqueFileName) {
        this.uniqueFileName = uniqueFileName;
    }

    public boolean isOverwrite() {
        return overwrite;
    }

    public void setOverwrite(boolean overwrite) {
        this.overwrite = overwrite;
    }

    public String getUploadSuffix() {
        return uploadSuffix;
    }

    /**
     * Set the file name suffix used during upload.  The suffix will be automatically removed as soon as the upload has completed.  
     * This allows other processes to discern completed files from files that are being uploaded.
     * 
     * @param uploadSuffix
     */
    public void setUploadSuffix(String uploadSuffix) {
        this.uploadSuffix = uploadSuffix;
    }

    // Implementation methods
    //-------------------------------------------------------------------------

    protected FTPClientPool createClientPool() throws Exception {
        FTPClientPool pool = new FTPClientPool();
        pool.afterPropertiesSet();
        return pool;
    }

    protected FTPClient borrowClient() throws Exception {
        try {
            return (FTPClient) getClientPool().borrowClient();
        } catch (Exception e) {
            throw new Exception(e);
        }
    }

    protected void returnClient(FTPClient client) {
        if (client != null) {
            try {
                getClientPool().returnClient(client);
            } catch (Exception e) {
                log.error("Failed to return client to pool: " + e, e);
            }
        }
    }
    
    @Override
    public void start() throws Exception {
    	super.start();
    }


	@Override
	public void stop() throws Exception {
		super.stop();
	}

	
	@Override
	public boolean run(Exchange exchange, Message inMessage, Message outMessage)
			throws Exception {
		
		FTPClient client = null;
	    try {
	        client = borrowClient();
	        // Change to the directory specified by the URI path if any
	        if (uri != null && uri.getPath() != null) {
	        	if (!client.changeWorkingDirectory(uri.getPath())) {
	        		log.warn("Unable to change ftp directory to '" + uri.getPath() + "'");
	            }
	        }
	        
		    OutputStream out = null;
		    String uploadName = null;
		    String fileName = null;
	        try {
		        List<String> names = getMarshaler().getOutputNames(exchange, inMessage);
		        for (Iterator<String> it = names.iterator(); it.hasNext();) {
		        	fileName = it.next();
			        if(fileName == null) {
	//		        	if(uniqueFileName != null)
	//		        		out = client.storeUniqueFileStream(uniqueFileName);
	//		        	else 
	//		        	out = client.storeUniqueFileStream();
			    	    UUID id = UUID.randomUUID();
			        	if(uniqueFileName != null)
			        		uploadName = uploadSuffix == null ? uniqueFileName + id.toString() : uniqueFileName + id.toString() + uploadSuffix;
			        	else 
			        		uploadName = uploadSuffix == null ? id.toString() : id.toString() + uploadSuffix;
			        	out = client.storeFileStream(uploadName);
			        } else {
			        	if(checkDuplicates && client.listFiles(fileName).length > 0) {
			        		if(overwrite) 
			        			client.deleteFile(fileName);
			                else
			                	throw new IOException("Can not send " + fileName + " : file already exists and overwrite has not been enabled");
			            }
			            uploadName = uploadSuffix == null ? fileName : fileName + uploadSuffix;
			            out = client.storeFileStream(uploadName);
			        }
			        if (out == null)
			        	throw new IOException("No output stream available for output name: " + uploadName + ". Maybe the file already exists?");
			       
			        getMarshaler().writeMessage(exchange, inMessage, fileName, out, uploadName);
		        }
		        
			} finally {
				if (out != null) {
					try {
						out.close();
		                client.completePendingCommand();
		                if(fileName != null && !fileName.equals(uploadName) && !client.rename(uploadName, fileName)) {
		                	throw new IOException("File " + uploadName + " could not be renamed to " + fileName);
		                }
		             } catch (IOException e) {
		            	 log.error("Caught exception while closing stream on error: " + e, e);
		             }
		        }
		        returnClient(client);
		    }
	        
	        
	    }catch (Exception e) {
	    	log.error("Error: ", e);
		}
		return true;
	}

}
