package org.spagic3.connectors.ftp;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.locks.Lock;

import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPFile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.ftp.locks.LockManager;
import org.spagic3.connectors.ftp.locks.SimpleLockManager;
import org.spagic3.connectors.ftp.temporaryData.FtpData;
import org.spagic3.core.routing.IMessageRouter;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;
import org.spagic3.nmr.api.Status;

public class PollingTask extends TimerTask {
	
	protected Logger log = LoggerFactory.getLogger(FtpPollerComponent.class);
	
	private IFtpReader poller;
	private LockManager lockManager;
	private ConcurrentMap<String, FtpData> openExchanges = new ConcurrentHashMap<String, FtpData>();
	private boolean messageSent;
	private String target = null;
	
	@Override
	public void run() {
		// Check if the target service is available
		try {
	        IMessageRouter router = poller.getMessageRouter();
	        if (!router.isServiceRegistered(this.target)){
	        	log.warn("Target Service ["+this.target+"] is not Registered, cannot fire message");
	        	return;
	        }
		} catch (Throwable th) {
        	log.error("Error during polling ", th);
        	return;
		}		
		
		if (lockManager == null) {
            lockManager = createLockManager();
        }
		try{
			pollFileOrDirectory(poller.getWorkingPath());
		}catch (Exception e) {
			log.error("Error while polling directory: " + poller.getWorkingPath(), e);
		}
	}
	
	 protected void pollFileOrDirectory(String fileOrDirectory) throws Exception {
	        FTPClient ftp = borrowClient();
	        try {
	            log.debug("Polling directory " + fileOrDirectory);
	            pollFileOrDirectory(ftp, fileOrDirectory, poller.isRecursive());
	        } finally {
	            returnClient(ftp);
	        }
	    }
	 
	   protected void pollFileOrDirectory(FTPClient ftp, String fileOrDirectory, boolean processDir) throws Exception {
		   FTPFile[] files = listFiles(ftp, fileOrDirectory);
		   for (int i = 0; i < files.length; i++) {
	       String name = files[i].getName();
	       if (".".equals(name) || "..".equals(name)) {
	           continue; // ignore "." and ".."
	       }
	       String file = fileOrDirectory + "/" + name;
	       // This is a file, process it
	       if (!files[i].isDirectory()) {
	           if (poller.getFilter() == null || poller.getFilter().accept(new File(file))) {
	              pollFile(file); // process the file
	           }
	           // Only process directories if processDir is true
	       } else if (processDir) {
	           if (log.isDebugEnabled()) {
	           	log.debug("Polling directory " + file);
	           }
	           pollFileOrDirectory(ftp, file, poller.isRecursive());
	       } else {
	           if (log.isDebugEnabled()) {
	           	log.debug("Skipping directory " + file);
	           }
	       }
	   }
	}
	   
	   protected void pollFile(final String file) {
	        if (log.isDebugEnabled()) {
	            log.debug("Scheduling file " + file + " for processing");
	        }
	        final Lock lock = lockManager.getLock(file);
	        if (lock.tryLock())
	        	processFileNow(file);
	        
	    }
	
	   /**
     * unlock the file
     *
     * @param file      the file to unlock
     */
    private void unlockAsyncFile(String file) {
        // finally remove the file from the open exchanges list
        Lock lock = lockManager.getLock(file);
        if (lock != null) {
            try {
                lock.unlock();
            } catch (Exception ex) {
                // can't release the lock
                log.error("Error trying to release the lock ", ex);
            }
        }
    }
    
    protected FTPClient borrowClient() throws Exception {
        try {
            return (FTPClient) poller.getClientPool().borrowClient();
        } catch (Exception e) {
            throw new Exception(e);
        }
    }
	
	 protected void processFileNow(String file) {
	        FTPClient ftp = null;
	        try {
	            ftp = borrowClient();
	            if (log.isDebugEnabled()) {
	            	log.debug("Processing file " + file);
	            }
	            if (listFiles(ftp, file).length > 0) {
	                // Process the file. If processing fails, an exception should be thrown.
	                processFile(ftp, file);
	                ftp = null;
	            } else {
	                //avoid processing files that have been deleted on the server
	            	log.debug("Skipping " + file + ": the file no longer exists on the server");
	            }
	        } catch (Exception e) {
	        	log.error("Failed to process file: " + file + ". Reason: " + e, e);
	        } finally {
	            if (ftp != null) {
	                returnClient(ftp);
	            }
	        }
	    }
	 
	   protected void returnClient(FTPClient client) {
	        if (client != null) {
	            try {
	                poller.getClientPool().returnClient(client);
	            } catch (Exception e) {
	                log.error("Failed to return client to pool: " + e, e);
	            }
	        }
	    }
	 
	    private FTPFile[] listFiles(FTPClient ftp, String directory) throws IOException {
	        if (poller.isChangeWorkingDirectory()) {
	            ftp.changeWorkingDirectory(directory);
	            return ftp.listFiles();
	        } else {
	            return ftp.listFiles(directory);
	        }
	    }
	 
	 protected void processFile(FTPClient ftp, String file) throws Exception {
        InputStream in = ftp.retrieveFileStream(file);
        Exchange exchange = poller.getFtpExchange();
        Message message = poller.getMessage(exchange);

        poller.getMarshaler().readMessage(exchange, message, in, file);
        if (poller.isStateless()) {
            exchange.setProperty(FtpData.class.getName(), new FtpData(file, ftp, in));
        } else {
            this.openExchanges.put(exchange.getId(), new FtpData(file, ftp, in));
        }
        
        poller.sendFtpExchange(exchange);
        messageSent = true;
    }
	
	public PollingTask(IFtpReader poller, String target) throws Exception {
		this.poller = poller;
		this.target = target;
		messageSent = false;
	}
	
    protected LockManager createLockManager() {
        return new SimpleLockManager();
    }
    
    public void process(Exchange exchange) throws Exception {
        FtpData data;
        if (poller.isStateless()) {
            data = (FtpData) exchange.getProperty(FtpData.class.getName());
        } else {
            data = this.openExchanges.remove(exchange.getId());
        }
        // check for done or error
        if (data != null) {
            log.debug("Releasing " + data.file);
            try {
                // Close ftp related stuff
                data.in.close();
                data.ftp.completePendingCommand();
                // check for state
                if (exchange.getStatus() == Status.Done) {
                    if (poller.isDeleteFile()) {
                        if (!data.ftp.deleteFile(data.file)) {
                            throw new IOException("Could not delete file " + data.file);
                        }
                    }
                } else {
                    Exception e = exchange.getError();
                    if (e == null) {
                        e = new Exception("Unkown error");
                    }
                    throw e;
                }
            } finally {
                // unlock the file
                unlockAsyncFile(data.file);
                // release ftp client
                returnClient(data.ftp);
            }
        } else {
            // strange, we don't know this exchange
            log.debug("Received unknown exchange. Will be ignored...");
        }
    }

	public boolean isMessageSent() {
		return messageSent;
	}

}
