/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.ftp.marshaler;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;

import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;

/**
 * A FileMarshaler that converts the given input stream into a binary
 * attachment.
 * 
 * @author Rossato Luca
 * @since 3.0
 */
public class BinaryFileMarshaler implements FileMarshaler {

	@Override
	public String getMarshalerId() {
		return "BINARY";
	}

    @Override
	public List<String> getOutputNames(Exchange exchange, Message message) {
    	Map<String, Object> attachments = message.getAttachments();
    	return Arrays.asList(attachments.keySet().toArray(new String[0]));
	}

	@Override
	public void writeMessage(Exchange exchange, Message message,
			String outputName, OutputStream out, String path)
			throws IOException {
		DataHandler handler = (DataHandler) message.getAttachment(outputName);
		if (handler == null) {
			throw new IOException("Could not find attachment: " + outputName);
		}
		InputStream is = handler.getInputStream();
		copyInputStream(is, out);
	}

	
	public void readMessage(Exchange exchange, Message message, InputStream in, String path)
        throws IOException {
        File polledFile = new File(path);
        
        // Put the attachment on a temporary file
        File tempFile = File.createTempFile("spagic-attachment", ".tmp" );

		OutputStream out = new FileOutputStream(tempFile);
		byte buf[] = new byte[1024];
		int len;
		while ((len = in.read(buf)) > 0)
			out.write(buf, 0, len);
		out.close();
		in.close();
		
        DataHandler handler = new DataHandler(new FileDataSource(tempFile));
        
        message.addAttachment(polledFile.getName(), handler);
        message.setHeader(FILE_NAME_PROPERTY, polledFile.getName());
        message.setHeader(FILE_PATH_PROPERTY, path);
        
        // Produce a body to avoid an output empty message
        message.setBody("<FILE>" + polledFile.getName() + "</FILE>");
    }

    /**
     * Copy in stream to an out stream
     *
     * @param in
     * @param out
     * @throws IOException
     */
    protected static void copyInputStream(InputStream in, OutputStream out) throws IOException {
        byte[] buffer = new byte[8192];
        int len = in.read(buffer);
        while (len >= 0) {
            out.write(buffer, 0, len);
            len = in.read(buffer);
        }
        in.close();
        out.close();
    }

}
