/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.ftp.marshaler;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.List;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.dom4j.DocumentHelper;
import org.dom4j.io.DocumentSource;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;

/**
 * A default file transformer which assumes the file is already in XML format and
 * requires no transformation other than to be wrapped in a normalized message..
 *
 * @org.apache.xbean.XBean
 * @version $Revision: 673769 $
 */
public class DefaultFileMarshaler implements FileMarshaler {

    private String encoding;

	@Override
	public String getMarshalerId() {
		return "DEFAULT";
	}

	public void readMessage(Exchange exchange, Message message, InputStream in, String path) throws IOException{
        if (encoding == null) {
            message.setBody(new StreamSource(in, path));
        } else {
            message.setBody(new StreamSource(new InputStreamReader(in, Charset.forName(encoding)), path));
        }
        message.setHeader(FILE_NAME_PROPERTY, new File(path).getName());
        message.setHeader(FILE_PATH_PROPERTY, path);
    }


    public List<String> getOutputNames(Exchange exchange, Message message) {
        return Arrays.asList(new String[]{asString(message.getHeader(FILE_NAME_PROPERTY))});
    }

    /**
     * creates a temporary file name for the given exchange when reading an inbound
     * message. This name is a (hopefully) unique name used for doing a first writing.
     * It will be renamed to the real output name after successfully finished transmission.
     * 
     * @param exchange the inbound message exchange
     * @param message the inbound message
     * @return the file name or null if a file name could not be found or calculated
     */
	public String getTempOutputName(Exchange exchange, Message message){
		Object retVal = message.getHeader(TEMP_FILE_NAME_PROPERTY);
		return retVal == null ? null : asString(retVal);
	}
	
	 /**
     * Converts the value to a String
     *
     * @param value the value to be coerced into a string
     * @return the value as a string or null if it cannot be converted
     */
    protected String asString(Object value) {
        return value != null ? value.toString() : null;
    }
    
    
    public void writeMessage(Exchange exchange, Message message, String outputName, OutputStream out, String path) throws IOException {
        try {
        	Object value = message.getBody();
            if (value != null) {
                writeValue(value, out);
            } else {
                writeMessageContent(exchange, message, out, path);
            }
        } catch (IOException e) {
            throw new IOException(e);
        }
    }

    // Properties
    //-------------------------------------------------------------------------   
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }
    
    public String getEncoding() {
        return encoding;
    }

    // Implementation methods
    //-------------------------------------------------------------------------

    /**
     * Writes the given value to the output stream
     *
     * @param value the value to be written to the stream
     * @param out   the output stream
     */
    protected void writeValue(Object value, OutputStream out) throws IOException {
        if (value instanceof String) {
            OutputStreamWriter writer = new OutputStreamWriter(out);
            writer.write((String) value);
            writer.flush();
        } else if (value instanceof byte[]) {
            out.write((byte[]) value);
        } else {
            ObjectOutputStream objectOut = new ObjectOutputStream(out);
            objectOut.writeObject(value);
        }
    }

    /**
     * Writes the message content to the given output stream
     *
     * @param message the message who's content we are about to write
     * @param out     the destination of the output
     * @param path    the name of the output resource (file, uri, url)
     */
    protected void writeMessageContent(Exchange exchange, Message message, OutputStream out, String path) throws IOException{
    	Object content = message.getBody();
        if (content == null) {
            throw new IOException("No message content in the inbound message for message exchange: " + exchange);
        }
        try {
            toResult(content, new StreamResult(out), encoding);
        } catch (TransformerException e) {
            throw new IOException(e);
        }
    }
    
    public void toResult(Object source, Result result, String charset) throws TransformerConfigurationException, TransformerException {
    	String defaultCharset = System.getProperty("org.spagic3.default.charset", "UTF-8");
    	if (source == null) {
	        return;
	    }
    	if (charset == null) {
            charset = defaultCharset;
        }
    	TransformerFactory factory = TransformerFactory.newInstance();
	    Transformer transformer = factory.newTransformer();
	    if (transformer == null) {
	        throw new TransformerException("Could not create a transformer - JAXP is misconfigured!");
	    }
	    transformer.setOutputProperty(OutputKeys.ENCODING, charset);
	    try{
	    	org.dom4j.Document inDocument = DocumentHelper.parseText((String)source);
			DocumentSource docSource = new DocumentSource(inDocument);
		    transformer.transform(docSource, result);
	    }catch (Exception e) {
			throw new TransformerException(e);
		}
	}
    
}
