/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.jms.marshaler;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringWriter;
import java.util.Map;

import javax.activation.DataHandler;
import javax.jms.Message;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamSource;

import org.spagic3.nmr.api.Exchange;
import org.spagic3.soap.core.MessageImpl;
import org.spagic3.soap.core.PhaseInterceptorChain;
import org.spagic3.soap.interceptors.mime.AttachmentsInInterceptor;
import org.spagic3.soap.interceptors.mime.AttachmentsOutInterceptor;
import org.spagic3.soap.interceptors.xml.BodyOutInterceptor;
import org.spagic3.soap.interceptors.xml.StaxInInterceptor;
import org.spagic3.soap.interceptors.xml.StaxOutInterceptor;

public class DefaultProviderMarshaler extends AbstractJmsMarshaler implements
    JmsProviderMarshaler {

    private Map<String, Object> jmsProperties;

    /**
     * @return the jmsProperties
     */
    public Map<String, Object> getJmsProperties() {
        return jmsProperties;
    }

    /**
     * @param jmsProperties the jmsProperties to set
     */
    public void setJmsProperties(Map<String, Object> jmsProperties) {
        this.jmsProperties = jmsProperties;
    }

    public Message createMessage(Exchange exchange, org.spagic3.nmr.api.Message in, Session session) throws Exception {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PhaseInterceptorChain chain = new PhaseInterceptorChain();
        chain.add(new AttachmentsOutInterceptor());
        chain.add(new StaxOutInterceptor());
        chain.add(new BodyOutInterceptor());
        org.spagic3.soap.api.Message msg = new MessageImpl();        
        msg.setContent(Source.class, new StreamSource(new java.io.StringReader((String)in.getBody())));
        
        msg.setContent(OutputStream.class, baos);
        for (String attId : in.getAttachments().keySet()) {
            msg.getAttachments().put(attId, (DataHandler) in.getAttachment(attId));
        }
        chain.doIntercept(msg);
        TextMessage text = session.createTextMessage(baos.toString());
        if (msg.get(org.spagic3.soap.api.Message.CONTENT_TYPE) != null) {
            text.setStringProperty(CONTENT_TYPE_PROPERTY,
                                   (String) msg.get(org.spagic3.soap.api.Message.CONTENT_TYPE));
        }
        if (jmsProperties != null) {
            for (Map.Entry<String, Object> e : jmsProperties.entrySet()) {
                text.setObjectProperty(e.getKey(), e.getValue());
            }
        }
        
        // handle correlation ID
        String correlationId = (String) in.getHeader("org.spagic3.components.jms.correlationId");
        if (correlationId != null) {
        	text.setJMSCorrelationID(correlationId);
        }

        if (isCopyProperties()) {
            copyPropertiesFromNM(in, text);
        }

        return text;
    }

    public void populateMessage(Message message, Exchange exchange, org.spagic3.nmr.api.Message normalizedMessage) throws Exception {
        if (message instanceof TextMessage) {
            PhaseInterceptorChain chain = new PhaseInterceptorChain();
            chain.add(new AttachmentsInInterceptor());
            chain.add(new StaxInInterceptor());
            org.spagic3.soap.api.Message msg = new MessageImpl();
            msg.setContent(InputStream.class, new ByteArrayInputStream(((TextMessage) message).getText().getBytes()));
            if (message.propertyExists(CONTENT_TYPE_PROPERTY)) {
                msg.put(org.spagic3.soap.api.Message.CONTENT_TYPE, message.getStringProperty(CONTENT_TYPE_PROPERTY));
            }
            chain.doIntercept(msg);
            XMLStreamReader xmlReader = msg.getContent(XMLStreamReader.class);
            javax.xml.transform.stax.StAXSource src = new javax.xml.transform.stax.StAXSource(xmlReader);
            
            javax.xml.transform.TransformerFactory tfactory = TransformerFactory.newInstance();
            javax.xml.transform.Transformer xform = tfactory.newTransformer();
            java.io.StringWriter writer = new StringWriter();
            javax.xml.transform.Result result = new javax.xml.transform.stream.StreamResult(writer);
            xform.transform(src, result);

            normalizedMessage.setBody(writer.toString());
            
            for (Map.Entry<String, DataHandler> attachment : msg.getAttachments().entrySet()) {
                normalizedMessage.addAttachment(attachment.getKey(), attachment.getValue());
            }
            if (isCopyProperties()) {
                copyPropertiesFromJMS(message, normalizedMessage);
            }
        } else {
            throw new UnsupportedOperationException("JMS message is not a TextMessage");
        }
    }

}
