/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.spagic3.connectors.mail;

import java.lang.reflect.Constructor;
import java.util.Date;
import java.util.Timer;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.mail.marshaler.AbstractMailMarshaler;
import org.spagic3.connectors.mail.marshaler.DefaultMailMarshaler;
import org.spagic3.connectors.mail.utils.MailConnectionConfiguration;
import org.spagic3.core.AbstractSpagicConnector;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Status;

/**
 * 
 * @author Luca Rossato
 */
public class MailPollerComponent extends AbstractSpagicConnector {
	protected Logger log = LoggerFactory.getLogger(MailPollerComponent.class);

	public static final String PROPERTY_MARSHALER = "marshaler";
	public static final String PROPERTY_CUSTOMTRUSTMANAGER = "customTrustManagers";
	public static final String PROPERTY_MAXFETCHSIZE = "maxFetchSize";
	public static final String PROPERTY_PROCESSONLYUNSEENMESSAGES = "processOnlyUnseenMessages";
	public static final String PROPERTY_DELETEPROCESSEDMESSAGES = "deleteProcessedMessages";
	public static final String PROPERTY_DEBUGMODE = "debugMode";
	public static final String PROPERTY_STARTTLS = "startTLS";
	public static final String PROPERTY_PERIOD = "period";
    
    // Container for polling parameters
	private MailParameters mailParams;
    
    // Status of the component: necessary because the method poll is called
	private boolean active; 
    private Timer timerService;
    
    public void init() {
		mailParams = new MailParameters();
		AbstractMailMarshaler abstractMailMarshaler;
			
		if(propertyConfigurator.getString(PROPERTY_MARSHALER)!=null && !propertyConfigurator.getString(PROPERTY_MARSHALER).equals("")){
			try{
			 Class classe = Class.forName(propertyConfigurator.getString(PROPERTY_MARSHALER));
			 Constructor c = classe.getConstructor(new Class[0]);
			 abstractMailMarshaler = (AbstractMailMarshaler)c.newInstance(new Object[0]);
			}catch(Exception e){
				log.error("The selected mail Marshaler is invalid: " + propertyConfigurator.getString(PROPERTY_MARSHALER) +
						" the defaul one will be used ", e);
				abstractMailMarshaler= new DefaultMailMarshaler();
			}
		}
		else
			abstractMailMarshaler = new DefaultMailMarshaler();		
		setMarshaler(abstractMailMarshaler);
		
		String host = propertyConfigurator.getString("host");
		String password = propertyConfigurator.getString("password");
		String protocol = propertyConfigurator.getString("protocol");
		String username = propertyConfigurator.getString("username");
		int port = propertyConfigurator.getInteger("port", -1);
		boolean startTLS = propertyConfigurator.getBoolean("startTLS",false);		
		String folderName = propertyConfigurator.getString("folderName","INBOX");
		
		MailConnectionConfiguration config = new MailConnectionConfiguration(host,password,protocol,username,startTLS,folderName,port);
		setMailConnectionConfiguration(config);
		setCustomTrustManagers(propertyConfigurator.getString(PROPERTY_CUSTOMTRUSTMANAGER,null));
		setConnection(config);
		setMaxFetchSize(propertyConfigurator.getInteger(PROPERTY_MAXFETCHSIZE, 5));
		
		setPeriod(propertyConfigurator.getLong(PROPERTY_PERIOD,(long)10000));
		
		
		setProcessOnlyUnseenMessages(propertyConfigurator.getBoolean(PROPERTY_PROCESSONLYUNSEENMESSAGES, true));
		setDeleteProcessedMessages(propertyConfigurator.getBoolean(PROPERTY_DELETEPROCESSEDMESSAGES, false));
		setDebugMode(propertyConfigurator.getBoolean(PROPERTY_DEBUGMODE, false));
		setStartTLS(propertyConfigurator.getBoolean(PROPERTY_STARTTLS, false));
				
	}

	@Override
	public void start() throws Exception {
		super.start();
		// Activate polling thread
		timerService = new Timer();
		timerService.schedule(new PollingTask(mailParams, this), new Date(), getPeriod());		

		// Component activated
		active = true;
		
	}

	@Override
	public void stop() throws Exception {
		super.stop();
		// Component deactivated
		active = false;
		timerService.cancel();
		timerService = null;
		
	}
	
	public boolean isActive() {
		return active;
	}

	public long getPeriod() {
		return mailParams.period;
	}
	
	public void setPeriod(long period) {
    	mailParams.period = period;
    }

    /**
     * @return the deleteProcessedMessages
     */
    public boolean isDeleteProcessedMessages() {
        return mailParams.deleteProcessedMessages;
    }

    /**
     * @param deleteProcessedMessages
     *            the deleteProcessedMessages to set
     */
    public void setDeleteProcessedMessages(boolean deleteProcessedMessages) {
    	mailParams.deleteProcessedMessages = deleteProcessedMessages;
    }

    /**
     * @return the marshaler
     */
    public AbstractMailMarshaler getMarshaler() {
        return mailParams.marshaler;
    }

    /**
     * @param marshaler
     *            the marshaler to set
     */
    public void setMarshaler(AbstractMailMarshaler marshaler) {
    	mailParams.marshaler = marshaler;
    }

    /**
     * @return the maxFetchSize
     */
    public int getMaxFetchSize() {
        return mailParams.maxFetchSize;
    }

    /**
     * @param maxFetchSize
     *            the maxFetchSize to set
     */
    public void setMaxFetchSize(int maxFetchSize) {
    	mailParams.maxFetchSize = maxFetchSize;
    }

    /**
     * @return the processOnlyUnseenMessages
     */
    public boolean isProcessOnlyUnseenMessages() {
        return mailParams.processOnlyUnseenMessages;
    }

    /**
     * @param processOnlyUnseenMessages
     *            the processOnlyUnseenMessages to set
     */
    public void setProcessOnlyUnseenMessages(boolean processOnlyUnseenMessages) {
        mailParams.processOnlyUnseenMessages = processOnlyUnseenMessages;
    }

    /**
     * returns the connection uri used for this poller endpoint
     * 
     * @return Returns the connection.
     */
    public String getConnection() {
        return mailParams.connection;
    }

    /**
     * sets the connection uri
     * 
     * @param connection
     *            The connection to set.
     */
    public void setConnection(MailConnectionConfiguration config) {
//    	<protocol>://<user>@<host>[:<port>][/<folder>]?password=<password> 
//    	<protocol>://<host>[:<port>][/<folder>]?user=<user>;password=<password>
    	String connStr = config.getProtocol()+"://"+config.getHost();
    	connStr += config.getPort()!= -1 ? ":"+config.getPort() : "";
    	connStr += config.getFolderName()!= "" ? "/"+config.getFolderName() : "";
    	connStr +="?user="+config.getUsername()+";password="+config.getPassword();
    	mailParams.connection = connStr;
	}
    
    public void setMailConnectionConfiguration(MailConnectionConfiguration config){
    	mailParams.config = config;
    }
    
//    public MailConnectionConfiguration getMailConnectionConfiguration() {
//        return mailParams.config;
//    }

    /**
     * @return the debugMode
     */
    public boolean isDebugMode() {
        return mailParams.debugMode;
    }

    /**
     * @param debugMode
     *            the debugMode to set
     */
    public void setDebugMode(boolean debugMode) {
    	mailParams.debugMode = debugMode;
    }

    /**
     * @return the customTrustManagers
     */
    public String getCustomTrustManagers() {
        return mailParams.customTrustManagers;
    }

    /**
     * @param customTrustManagers
     *            the customTrustManagers to set
     */
    public void setCustomTrustManagers(String customTrustManagers) {
    	mailParams.customTrustManagers = customTrustManagers;
    }

	public boolean isStartTLS() {
		return mailParams.startTLS;
	}

	public void setStartTLS(boolean startTLS) {
		mailParams.startTLS = startTLS;
	}	
	
	@Override
	public void process(Exchange exchange) {
		if(!exchange.getStatus().equals(Status.Done))
			throw new IllegalStateException("Error unexppected message status: "+ exchange.getStatus().name());
	}
	
	public Exchange getExchange(){
		return createInOnlyExchange();
	}
	
	/**
	 * Send a message to the NMR
	 * 
	 * @param messageStringBuffer
	 * @throws Exception
	 */
	protected void sendMessage(Exchange exchange){
		log.debug("Sending message ");
		
		send(exchange);
	}
}
