package org.spagic3.connectors.mail;

import java.util.Properties;
import java.util.TimerTask;

import javax.mail.Flags;
import javax.mail.Folder;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Store;
import javax.mail.internet.MimeMessage;
import javax.mail.search.FlagTerm;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.mail.utils.MailUtils;
import org.spagic3.nmr.api.Exchange;


public class PollingTask extends TimerTask {
	
	protected Logger log = LoggerFactory.getLogger(MailPollerComponent.class);
	
	private MailParameters mailParams;
	private MailPollerComponent poller;

	@Override
	public void run() {
		log.debug("Polling mailfolder " + mailParams.config.getFolderName() + " at host " + mailParams.config.getHost() + "...");

        if (poller.getMaxFetchSize() == 0) {
        	log.debug("The configuration is set to poll no new messages at all...skipping.");
            return;
        }

        Store store = null;
        Folder folder = null;
        Session session = null;
        try {
        	mailParams.config.setStartTLS(poller.isStartTLS());
            Properties props = MailUtils.getPropertiesForProtocol(mailParams.config, poller.getCustomTrustManagers());
            props.put("mail.debug", poller.isDebugMode() ? "true" : "false");

            // Get session
            session = Session.getInstance(props, mailParams.config.getAuthenticator());

            // debug the session
            session.setDebug(poller.isDebugMode());
            Message[] messages = null;
            try{
	            store = session.getStore(mailParams.config.getProtocol());
	            store.connect(mailParams.config.getHost(), mailParams.config.getUsername(), mailParams.config.getPassword().toString());
	            folder = store.getFolder(mailParams.config.getFolderName());
	            if (folder == null || !folder.exists()) {
	                throw new Exception("Folder not found or invalid: " + mailParams.config.getFolderName());
	            }
	            folder.open(Folder.READ_WRITE);
	
	            if (poller.isProcessOnlyUnseenMessages()) {
	                messages = folder.search(new FlagTerm(new Flags(Flags.Flag.SEEN), false));
	            } else {
	                messages = folder.getMessages();
	            }
            }catch (Exception e) {
            	log.error("Error retrieving Mail messages ", e);
			}    
            
            //StringBuffer collectionMessageString = new StringBuffer();

            int fetchSize = poller.getMaxFetchSize() == -1 ? messages.length : Math.min(poller.getMaxFetchSize(), messages.length);
            for (int cnt = 0; cnt < fetchSize; cnt++) {
                // get the message
                MimeMessage mailMsg = (MimeMessage)messages[cnt];
                
                // create the exchange
                Exchange exchange = poller.getExchange();
                
                org.spagic3.nmr.api.Message inMessage = exchange.getIn(true);
                
                // now let the marshaller convert the mail into a normalized
                // message to send to jbi bus
                try{
                	poller.getMarshaler().convertMailToNormalizedMessage(exchange, inMessage, mailMsg);
                }catch (MessagingException me) {
                	log.error("Error converting Mail message ", me);
				}

                // Send message to NMR
    			if (exchange.getIn(false).getBody()!= null) {
    				poller.sendMessage(exchange);
    			}

//                // now check if delivery succeeded or went wrong
//                if (io.getStatus() == ExchangeStatus.ERROR) {
//                    Exception e = io.getError();
//                    if (e == null) {
//                        e = new JBIException("Unexpected error occured...");
//                    }
//                    throw e;
//                } else {
    			try{
                    // then mark the mail as processed (only if no errors)
                    if (poller.isDeleteProcessedMessages()) {
                        // processed messages have to be marked as deleted
                        mailMsg.setFlag(Flags.Flag.DELETED, true);
                    } else {
                        // processed messages have to be marked as seen
                        mailMsg.setFlag(Flags.Flag.SEEN, true);
                    }
    			}catch (MessagingException me) {
                	log.error("Error handling processed mail ", me);
				}
            }    
        } finally {
            // finally clean up and close the folder and store
            try {
                if (folder != null) {
                    folder.close(true);
                }
                if (store != null) {
                    store.close();
                }
            } catch (Exception ignored) {
                log.error("Error during polling ",ignored);
            }
        }		
	}
	
	public PollingTask(MailParameters mailParams, MailPollerComponent poller) throws Exception {
		this.mailParams = mailParams;
		this.poller = poller;
	}

}
