package org.spagic3.connectors.mail.custom.marshaller;

import java.util.Enumeration;

import javax.mail.Header;
import javax.mail.Message;
import javax.mail.Multipart;
import javax.mail.Part;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.mail.internet.MimePart;
import javax.mail.internet.MimePartDataSource;

import org.dom4j.Document;
import org.dom4j.io.SAXReader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>PECMimeMessage</p>
 * <p>
 * This class let you to manage an email message produce by a server conforming to the requirement provided
 * by CNIPA ( an Italian Organization ) for an email to be compliant with the standard called "PEC - Posta Elettronica Certificata"
 * 
 * The CNIPA Defines that a message conforming to this standard has the following structure
 * 
 * +--------------------------------+
 * |    MIME Message                |
 * +----- Part 0 -------------------+
 * |  +--------------------------+  |
 * |  |   MIME Message           |  |
 * |  +------ Part 0 ------------+  |
 * |  |       Body               |  |
 * |  +------ Part 1 ------------+  |
 * |  | Attachment: daticert.xml |  |
 * |  +--------------------------+  |
 * |  |   <MIME Message>         |  |
 * |  +--------------------------+  |
 * +------ Part 1 ------------------+
 * | Attachment: smime.p7s          |
 * +--------------------------------+
 *  *
 * </p>
 * <p>Copyright: Copyright (c) 2008</p>
 * <p>Company: Engineering Ing. Informatica S.p.A</p>
 * @author Giulio Toffoli
 * @author Andrea Zoppello
 * @version 1.0
 */

public class PECMimeMessage {

	protected Logger log = LoggerFactory.getLogger(PECMimeMessage.class);	
	
  public static final int ACCETTAZIONE = 1;
  public static final  int PRESA_IN_CARICO = 2;
  public static final int AVVENUTA_CONSEGNA = 3;
  public static final int POSTA_CERTIFICATA = 4;
  public static final int ERRORE_CONSEGNA = 5;
  public static final int ANOMALIA = 6;
  public static final int NESSUN_TIPO_SPECIFICATO = 0; // Rappresenta un errore...

  private javax.mail.Message pecMsg; // This is the message as it is originally received by PEC Server
  
  private String UIDL="";
  
  public PECMimeMessage(Message pecMsg) {
        this.pecMsg = pecMsg;
  }

 
  private int getPECMessageType() throws javax.mail.MessagingException {
	// Cerchiamolo tra gli headers del messaggio...

	Enumeration<Header> e1 = pecMsg.getAllHeaders();
    while (e1.hasMoreElements())
    {
      javax.mail.Header header = (javax.mail.Header)e1.nextElement();
      
      log.debug(header.getName() + ": "+ header.getValue());
      
      if (header.getName().startsWith("X-Ricevuta"))
      {
        if (header.getValue().equals("accettazione")) return this.ACCETTAZIONE;
        if (header.getValue().equals("presa-in-carico")) return this.PRESA_IN_CARICO;
        if (header.getValue().equals("avvenuta-consegna")) return this.AVVENUTA_CONSEGNA;
        if (header.getValue().equals("errore-consegna")) return this.ERRORE_CONSEGNA;
      }
      else if (header.getName().startsWith("X-Trasporto"))
      {
        if (header.getValue().equals("errore")) return this.ANOMALIA;
        if (header.getValue().equals("posta-certificata")) return this.POSTA_CERTIFICATA;
      }
    }

    return this.NESSUN_TIPO_SPECIFICATO;

  }

 
  /**
   * Retrieve The Original Message Within The PEC Message
   */
  public PECTransferObject parse() throws Exception {
    	PECTransferObject pecTO = new PECTransferObject(); 	
	  	try {
    		  log.debug("getOriginalMessage->start");
    		  log.debug("getOriginalMessage->Parse the PEC Multipart Message");
              
    		  Multipart multiPart = (Multipart)pecMsg.getContent();
    		  pecTO.setType(getPECMessageType());
              log.debug("getOriginalMessage-> Parsed ["+multiPart.getCount()+"] MIME Part");
              if (multiPart.getCount() != 2) {
                  throw new Exception("This MIME Message has ["+multiPart.getCount()+"] Parts, PEC Message must have exactly 2 Part");
              }

              log.debug("getOriginalMessage-> Parsing the 0 Part");
              Part bodyPart = multiPart.getBodyPart(0);
              log.debug("getOriginalMessage-> Part[0] Content Type + "+bodyPart.getContentType());

    	      if (bodyPart.getContentType() == null || !bodyPart.getContentType().startsWith("multipart/mixed;")) {
                  throw new Exception("Mime Part [0] must be of type multipart/mixed");
              }

    	      log.debug("getOriginalMessage-> Parsing the Body As Multipart Itself"); 
    	      // The Body Part is It self a Multipart
              MimePartDataSource multipartDataSource = new MimePartDataSource( (MimePart)bodyPart );
              MimeMultipart bodyMultiPart = new MimeMultipart(multipartDataSource);
              
              log.debug("getOriginalMessage-> bodyMultiPart is built, parsing it");

              Part partOfBody = null;
              String disposition = null;
    	      for (int idx=0; idx < bodyMultiPart.getCount(); idx++) {
    	    	 
    	    	 log.debug("getOriginalMessage-> Reading the ["+idx+"] Part of the Body");
    	    	 partOfBody = bodyMultiPart.getBodyPart(idx);
    	    	 disposition = partOfBody.getDisposition();
              	
              	 log.debug("getOriginalMessage-> Part ["+idx+"] has Disposition ["+disposition+"]");

              	 if ((disposition != null) && (disposition.equals(Part.ATTACHMENT) || disposition.equals(Part.INLINE))){
              		log.debug("getOriginalMessage-> Part ["+idx+"] Attached File [" + partOfBody.getFileName()+"]" );
              		if (partOfBody.getFileName() != null && partOfBody.getFileName().endsWith(".eml")){
                   		pecTO.setOriginalMessage(new MimeMessage(null, partOfBody.getInputStream()));
                    }else if (partOfBody.getFileName() != null && partOfBody.getFileName().endsWith("daticert.xml")){
                    	 // Parse datiCert.xml with DOM4J
                    	 SAXReader reader = new SAXReader(false);
                    	 Document document = reader.read(partOfBody.getInputStream());
                    	 pecTO.setDatiCert(document.asXML());
                    }
                }
              }
    	      return pecTO;
        } catch (Exception ex) {
        	log.error("Exception Occurred ", ex);
        	throw new Exception(" Not Valid PEC Message", ex);
        }
  }    
     /** Getter for property UIDL.
      * @return Value of property UIDL.
      *
      */
     public java.lang.String getUIDL() {
         return UIDL;
     }
     
     /** Setter for property UIDL.
      * @param UIDL New value of property UIDL.
      *
      */
     public void setUIDL(java.lang.String UIDL) {
         this.UIDL = UIDL;
     }
     
    

	public javax.mail.Message getPecMsg() {
		return pecMsg;
	}

	public void setPecMsg(javax.mail.Message pecMsg) {
		this.pecMsg = pecMsg;
	}
     
}
  




