/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic3.connectors.tcp;

import java.security.GeneralSecurityException;

import org.apache.commons.codec.binary.Base64;
import org.apache.mina.common.ByteBuffer;
import org.apache.mina.common.ConnectFuture;
import org.apache.mina.common.DefaultIoFilterChainBuilder;
import org.apache.mina.common.IoConnector;
import org.apache.mina.common.IoSession;
import org.apache.mina.common.RuntimeIOException;
import org.apache.mina.common.WriteFuture;
import org.apache.mina.filter.LoggingFilter;
import org.apache.mina.filter.SSLFilter;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.transport.socket.nio.SocketConnector;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.tcp.codec.util.NMUtils;
import org.spagic3.connectors.tcp.ssl.TCPBCSSLContextFactory;
import org.spagic3.core.AbstractSpagicConnector;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;
import org.spagic3.nmr.api.Pattern;
import org.spagic3.nmr.api.Status;

/**
 * Implement the "TCPClient" usin MINA asynchronous library. Based on
 * code of the Spagic 2.X TCP/IP provider endpoint by Gianfranco
 * Boccalon and contributions of Gian Uberto Lauri.
 * 
 * @author <a href="mailto:saint@eng.it">Gian Uberto Lauri</a>
 * @version $Revision$
 */
public class TCPClient extends AbstractSpagicConnector {

	private static final Logger log = LoggerFactory.getLogger(TCPClient.class);

	private static final double retryPower = 2.0;
	private TCPBCConfig config;
	private TCPMarshaller marshaller;
	private SpagicConnectFuture future;

	private FIFOMonitor fifoMonitor = new FIFOMonitor();

	@Override
	/**
	 * <code>init</code> Configures the component upon instantiation.
	 *
	 */
	public void init() {

		log.info("-- TCP Client Component Init --");

		TCPBCConfig config = new TCPBCConfig();

		config.setBase64decode(propertyConfigurator.getBoolean("base64.decode"));
		config.setBase64encode(propertyConfigurator.getBoolean("base64.encode"));
		config.setConnectionMode(propertyConfigurator.getString("connection.mode"));
		config.setInNmEnvelope(propertyConfigurator.getString("in.envelope"));
		config.setOutNmEnvelope(propertyConfigurator.getString("out.envelope"));
		config.setRemoteHost(propertyConfigurator.getString("remote.address"));
		config.setRemotePort(propertyConfigurator.getInteger("remote.port"));
		config.setRetryCount(propertyConfigurator.getInteger("retry.count"));
		config.setRetryDelay(propertyConfigurator.getLong("retry.delay"));
		config.setRetryNumber(propertyConfigurator.getInteger("retry.number"));
		config.setRetryType(propertyConfigurator.getString("retry.type"));
		config.setResponseTimeout(propertyConfigurator.getLong("reply.timeout"));
		config.setUseSSL(propertyConfigurator.getBoolean("connection.clientssl"));

		config.setIncomingHeader(propertyConfigurator.getString("incoming.header"));
		config.setIncomingTrailer(propertyConfigurator.getString("incoming.trailer"));
		config.setIncomingEndianness(propertyConfigurator.getString("incoming.endianness"));
		config.setOutgoingHeader(propertyConfigurator.getString("outgoing.header"));
		config.setOutgoingTrailer(propertyConfigurator.getString("outgoing.trailer"));
		config.setOutgoingEndianness(propertyConfigurator.getString("outgoing.endianness"));
		config.setStripWrapping(propertyConfigurator.getBoolean("strip.wrapping"));
		config.setTcpOutInReceiverClassName(propertyConfigurator.getString("tcp.inoutreceiver"));

		this.config = config;
	}


	@Override
	/**
	 * <code>start</code> Just logs that the component has been
	 * started
	 *
	 * @exception Exception if something goes berserk (should never
	 * happen, at least with the current implementation)
	 */
	public void start() throws Exception {
		super.start();
		log.info("-- TCP Client Component Start --");
	}


	@Override
	/**
	 * <code>stop</code> stops the component shutting down any pending
	 * communication.
	 *
	 * @exception Exception if something awkward happens...
	 */
	public void stop() throws Exception {
		super.stop();
		// close any pending asynchronous I/O
		if (future != null) {
			((TCPClientProtocolHandler) future.getSession().getHandler()).dispose();
			future.getSession().close();
		}
	}


	private TCPMarshaller getMarshaller() {
		if (marshaller == null) {
			marshaller = new TCPMarshaller(config);
		}
		return marshaller;
	}

	/**
	 * <code>getEndpoint</code> returns the ID of this endpoint
	 *
	 * @return a <code>String</code> the id of this endpoint
	 */
	protected String getEndpoint(){
		return this.getSpagicId();
	}

	@Override
	/**
	 * Sends the content of the in message to the remote host
	 * specified in the configuration -maybe overridden by the message
	 * contents- using a TCP/IP connection. For this pupose uses
	 * Apache MINA libraries for asynchronous I/O.
	 *
	 * @param exchange an <code>Exchange</code> the message excange that
	 * comes from the Service Bus, containing the input and output messages
	 */
	public void process(Exchange exchange) {
		Message in;
		Message out;

		if (exchange.getPattern() == Pattern.InOnly){
			in = exchange.getIn();
			out = null;
		}
		else {
			in = exchange.getIn();
			out = exchange.getOut();
		}

		// The method  could become  more complicated that  the simple
		// return this and that, so let's use it, but cache the result
		// for some efficency.
		String myEndpoint = getEndpoint();

		log.debug(myEndpoint+ 
				  "processInOut->start and processInOut->Creating IOConnector");
		IoConnector connector = new SocketConnector();

		log.debug(myEndpoint + 
				  "processInOut->IOConnector Creaing Filter Chain");
		DefaultIoFilterChainBuilder chain = connector.getDefaultConfig()
			.getFilterChain();

		if (config.isUseSSL()) {
			log.debug("processInOut->SSL Supporr Required configure SSL In Chain");
			try{
				addSSLSupport(chain);
			}catch (GeneralSecurityException e) {
				log.error(e.getMessage(), e);
				RuntimeException stone = 
					new RuntimeException("A GeneralSecurityException occurred: " 
							+ e.getMessage(), e);
				stone.setStackTrace(e.getStackTrace());
				throw stone;
			}
		}
		log.debug(myEndpoint + "processInOut-> Add The Codec Protocol Filter ");
		chain.addLast("codec", new ProtocolCodecFilter(new org.spagic3.connectors.tcp.codec.TCPMsgCodecFactory(config)));

		if (log.isDebugEnabled()) {
			addLogger(chain);
		}

		String threadId = myEndpoint + Thread.currentThread().getId();

		fifoMonitor.enterMonitor(threadId);
		try {
			log.debug(myEndpoint + "processInOut-> call sendTCPMessage : Entered in Synchnronized Code Block ");
			sendTCPMessage(connector, exchange, in, out);
			log.debug(myEndpoint + "processInOut-> after send : Finish Synchnronized Code Block ");
		}
		finally {
			fifoMonitor.leaveMonitor(threadId);

		}


	}

	private IoSession getTCPSession(ConnectFuture connectFuture) {
		// can throw exception
		IoSession aSession = null;
		long previousDelayTime = config.getRetryDelay();

		log.debug(getEndpoint() + "sendTCPMessage::getTCPSession Getting Session");

		log.debug(getEndpoint() + "sendTCPMessage::getTCPSession Future Is Ready -> [" + connectFuture.isReady() + "]");
		if (!connectFuture.isConnected()) {
			// log.debug(getEndpoint() + "sendTCPMessage::getTCPSession Future Is Not Connected
			// -> Must JOIN");
			// future.join();
			// log.debug(getEndpoint() + "JOIN OK");
		}

		// -> Ciclo For
		for (int retry = 0; retry <= config.getRetryNumber(); retry++) {
			log.debug("sendTCPMessage::getTCPSession :: Getting TCP Session Retry [" + retry + "] Getting session");
			aSession = connectFuture.getSession();

			if (aSession != null && !aSession.isClosing() && aSession.isConnected()) {
				log.debug("sendTCPMessage::getTCPSession :: Getting TCP Session Retry [" + retry + "] Getted OK ");
				break;
			} else {
				log.debug("sendTCPMessage::getTCPSession Session is NOT VALID ");
				if (retry <= config.getRetryNumber()) {
					long delay = getRetryWaitTime(previousDelayTime);
					log.debug("sendTCPMessage::getTCPSession:: Sleeping For Retry Wait Time" + delay);
					if (delay == -1) {
						aSession = null;
						break;
					}
					log.debug("sendTCPMessage::getTCPSession::Retring in: (ms) " + delay);
					previousDelayTime = delay;
					try {
						Thread.sleep(delay);
					} catch (InterruptedException ie) {
						log.error("Interrupted waiting for retry", ie);
					}
				}
			}
		}
		return aSession;

	}

	private void sendTCPMessage(IoConnector connector, Exchange exchange, 
								Message in, Message out) 
	throws RuntimeIOException {
		String myEndpoint = getEndpoint();

		try{
		log.debug(myEndpoint + ": sendTCPMessage(connector,exchange,in,out)");
		// connect to the TCPServer using Retry policies
		SpagicConnectFuture connectFuture = getConnection(connector, in);
		if (connectFuture == null) {
			// no connection is possible
			log.error(myEndpoint+": sendTCPMessage::connect future is null");
			throw new RuntimeIOException("sendTCPMessage::connect future is null");
		}

		IoSession session = getTCPSession(connectFuture);

		if (session != null && !session.isConnected()){
			// if session isn't connected create a new connection
			this.future = null;
			connectFuture = getConnection(connector, in);
			session = getTCPSession(connectFuture);
		}
		else if (session == null) {
			String remoteHost = getDynamicRemoteHost(in);
			Integer remotePort = getDynamicRemotePort(in);

			StringBuilder msg = new StringBuilder(getEndpoint());
			msg.append("Error creating connection host=");
			msg.append(remoteHost);
			msg.append(", port=");
			msg.append(remotePort);

			log.warn(msg.toString()); 
			future = null;
			throw new RuntimeIOException(msg.toString());

		}
		else if (session.isConnected()) {
			log.debug(getEndpoint() + "Session Is Connected");
			try {
				Object received = null;
				boolean isNullResponse = false;
				session.setAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY, Boolean.FALSE);
				int maxRetry = config.getRetryCount();
				for (int numRetry = 0; numRetry < maxRetry && received == null; numRetry++) {
					// write the message
					Object msgObj = NMUtils.retrieveMessageInEnvelope(in, config.getInNmEnvelope());
					if (config.isBase64decode()) {
						byte[] decodedB = Base64.decodeBase64((byte[]) msgObj);
						msgObj = decodedB;
					}

					log.debug(getEndpoint() + "sendTCPMessage::Retry=" + numRetry + " -> Message to send  " + msgObj);

					connectFuture.resetReply();
					WriteFuture wFuture = session.write(msgObj);
					wFuture.join();
					log.debug(getEndpoint() + "sendTCPMessage::Retry=" + numRetry + " -> After wFuture.join()");

					// This makes sense, there are 2 more kind of exchange pattern.
					if (exchange.getPattern() == Pattern.InOnly || exchange.getPattern() == Pattern.InOut) {
						// terminate the exchange
						exchange.setStatus(Status.Done);
					}

					try {
						log.debug(getEndpoint() + "sendTCPMessage::Retry=" + numRetry + " -> Waiting For Response ");
						received = connectFuture.waitForResponse(session);

						log.debug(getEndpoint() + "sendTCPMessage::Retry=" 
								  + numRetry+ " -> Response is " 
								  + ( received == null ? "Is Null" 
									  : "Not Null"));

						if (received!=null){
							if (received instanceof ByteBuffer) {
								log.debug(getEndpoint()+"sendTCPMessage::Retry=" + numRetry + " -> Transforming Response To Normalized Message ");
								// Out == null means that we have an in-only exchange, that is, fire and forget,
								//             do not care of putting the reply somewhere...
								if(out != null) {
									getMarshaller().toNMS(out, (ByteBuffer)received);
									log.info(getEndpoint() + " sendTCPMessage::Retry=" + numRetry+ " -> OK " + out.getBody(String.class));
								}
								else {
									log.info(getEndpoint() + " sendTCPMessage::Retry=" + numRetry+ " -> OK ");
								}
								session.setAttribute(TCPClientProtocolHandler.MESSAGE_INSESSION_KEY);
							}
							else {
								isNullResponse = true;
								log.warn(getEndpoint()+"sendTCPMessage::Retry=" + numRetry + " -> Null response, Unable to handle tcp response of type: " +
									 received.getClass().getName() );
							}
						} 
					} 
					catch (InterruptedException ie) {
						log.error(getEndpoint()+"sendTCPMessage::Retry=" + numRetry + " -> Error waiting for a response", ie);
					} 
					catch (Throwable t) {
						log.error(getEndpoint()+"sendTCPMessage::Retry=" + numRetry + " -> Uncatched Exception Occurred", t);
					} 
					finally {
						// prepare session for next request response exchange
						session.setAttribute(TCPClientProtocolHandler.WAIT_RESPONSE_KEY);
					}
				}

				if (received == null ) {
					log.warn(getEndpoint()+"sendTCPMessage::No response received after retry policy");

					Element faultElem = DocumentHelper.createElement("Fault");
					faultElem.setText("No response received from TCP server");
					log.error("No response received from TCP server");
					if(exchange.getFault() != null)
						exchange.getFault().setBody(faultElem.asXML());
					if(out != null)
						out.setBody(faultElem.asXML());
				}else if (isNullResponse){
					log.warn(getEndpoint()+"sendTCPMessage::Received a Null Response");
					Element faultElem = DocumentHelper.createElement("Fault");
					faultElem.setText("Received a Null Response From TCP Server");
					log.error("Received a Null Response From TCP Server");
					if(exchange.getFault() != null)
						exchange.getFault().setBody(faultElem.asXML());
					if(out != null)
						out.setBody(faultElem.asXML());
				}

			} finally {
				closeConnection(connectFuture);
			}
		} else {
			log.warn("sendTCPMessage::Session not created!");
		}
		}catch (Exception e) {
			throw new IllegalStateException(e.getMessage(), e);
		}

	}


	private void closeConnection(ConnectFuture future) {

		if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode()) 
				|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config.getConnectionMode())) {
			log.debug("closeConnection::Endpoint Configured To Mantain Session ");
			log.debug(getEndpoint() + "Nothing to do on future and session");
		} else {
			log.debug(getEndpoint() + "closeConnection::Endpoint Configured To Not Mantain Session ");
			log.debug(getEndpoint() + "closeConnection::Endpoint Configured To Not Mantain Session close session and future");
			future.getSession().close();
			future = null;
		}
	}

	private String getDynamicRemoteHost(Message in) {
		String remoteHost = (String) in.getHeader(TCPBCConfig.TCP_DESTINATION_HOST);
		if (remoteHost == null) {
			remoteHost = config.getRemoteHost();
		} else {
			log.debug(getEndpoint() + "Remote Host URI overridden: " + remoteHost);
		}
		return remoteHost;
	}

	private Integer getDynamicRemotePort(Message in) {

		Integer remotePort = (Integer) in.getHeader(TCPBCConfig.TCP_DESTINATION_PORT);
		if (remotePort == null) {
			remotePort = config.getRemotePort();
		} else {
			log.debug(getEndpoint() + "Remote port URI overridden: " + remotePort);
		}
		return remotePort;
	}

	/**
	 * Connect to the TCPServer using retry policies: <br/><br/> From Rapsody
	 * manual:<br/> The Retry Count setting determines how many times the
	 * message should be resent while waiting for a response. This can be set to
	 * 0, to indicate that the message should never be resent, or -1, to
	 * indicate that there is no maximum number of resends. Once the maximum
	 * number of resends has been reached and the timeout occurs, the Fail
	 * Action is executed (which is one of Send Message to Error Queue, Close
	 * Connection, or Send to Error Queue and Close Connection.
	 * 
	 * @param connector
	 * @return
	 */
	private SpagicConnectFuture getConnection(IoConnector connector,
			Message in) {
		long previousDelayTime = config.getRetryDelay();

		if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode())
				|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config.getConnectionMode())) {
			if ((future != null) && future.isConnected()) {
				log.debug(getEndpoint()+"getConnection::Returning already initiated connection");
				return future;
			}
		}

		for (int retry = 0; retry <= config.getRetryNumber(); retry++) {
			log.debug(getEndpoint() + "getConnection::Retry=" + retry + " start");

			try {
				String remoteHost = getDynamicRemoteHost(in);
				log.debug(getEndpoint() + "getConnection::Retry=" + retry + " Remote Host" + remoteHost);
				Integer remotePort = getDynamicRemotePort(in);
				log.debug(getEndpoint() + "getConnection::Retry=" + retry + " Remote Port" + remotePort);
				log.debug(getEndpoint() + "getConnection::Retry=" + retry + " Before Connect");
				SpagicConnectFuture localFuture = new SpagicConnectFuture(connector, config, remoteHost, remotePort);

				// connection obtained
				localFuture.join();
				log.debug(getEndpoint() + "getConnection::Retry=" + retry + " After Connect");

				if (TCPBCConfig.CONN_MODE_MAINTAIN.equals(config.getConnectionMode())
					|| TCPBCConfig.CONN_MODE_MANTAIN.equals(config.getConnectionMode())) {
					// I forget the previous instance????
					future = localFuture;
				}
				log.debug(getEndpoint() + "getConnection::Retry=" + retry + " Return local Future");
				return localFuture;
			} catch (Exception e) {
				log.error(getEndpoint()+"Error connecting to TCPServer", e);
				if (retry <= config.getRetryNumber()) {
					long delay = getRetryWaitTime(previousDelayTime);
					if (delay == -1) {
						return null;
					}
					log.debug(getEndpoint() + "...Retring in: (ms) " + delay);
					previousDelayTime = delay;
					try {
						Thread.sleep(delay);
					} catch (InterruptedException ie) {
						log.error(getEndpoint()+"Error waiting for retry", ie);
					}
				}
			}
		}
		return null;
	}

	private long getRetryWaitTime(long previousTime) {
		String retryType = config.getRetryType();
		if (retryType.equals(CommonConfig.RETRY_TYPE_IMMEDIATE)) {
			return 0;
		}
		if (retryType.equals(CommonConfig.RETRY_TYPE_LINEAR)) {
			return config.getRetryDelay();
		}
		if (retryType.equals(CommonConfig.RETRY_TYPE_EXPONENTIAL)) {
			return (long) Math.pow(previousTime, retryPower);
		}
		if (retryType.equals(CommonConfig.RETRY_TYPE_NO_RETRY)) {
			return -1;
		}

		return -1;
	}

	private void addSSLSupport(DefaultIoFilterChainBuilder chain) throws GeneralSecurityException {

			SSLFilter sslFilter = new SSLFilter(TCPBCSSLContextFactory.getInstance(config));
			sslFilter.setUseClientMode(true);
			chain.addLast("SSL", sslFilter);

		log.info("SSL ON");
	}

	public TCPBCConfig getConfig() {
		return config;
	}

	public void setConfig(TCPBCConfig config) {
		this.config = config;
	}

	private void addLogger(DefaultIoFilterChainBuilder chain)  {
		// we have to log CONNECTION and DATA.
		// TODO CHECK If the logger is added after the codec it receives data
		// already
		// converted to the codec protocol?
		chain.addLast("logger", new LoggingFilter());
		log.info("Logging ON");
	}


}
