/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic3.connectors.tcp;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLEngine;

import org.apache.commons.codec.binary.Base64;
import org.apache.mina.common.ByteBuffer;
import org.apache.mina.common.DefaultIoFilterChainBuilder;
import org.apache.mina.common.IoAcceptor;
import org.apache.mina.common.IoAcceptorConfig;
import org.apache.mina.common.IoSession;
import org.apache.mina.common.SimpleByteBufferAllocator;
import org.apache.mina.filter.LoggingFilter;
import org.apache.mina.filter.SSLFilter;
import org.apache.mina.filter.codec.ProtocolCodecFilter;
import org.apache.mina.transport.socket.nio.SocketAcceptor;
import org.apache.mina.transport.socket.nio.SocketAcceptorConfig;
import org.apache.mina.util.NewThreadExecutor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.connectors.tcp.codec.TCPMsgCodecFactory;
import org.spagic3.connectors.tcp.codec.util.NMUtils;
import org.spagic3.connectors.tcp.ssl.TCPBCSSLContextFactory;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.core.AbstractSpagicConnector;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;
import org.spagic3.nmr.api.Status;

public class TCPServer extends AbstractSpagicConnector {
	
	
	private static final String SSL_RSA_WITH_NULL_SHA = "SSL_RSA_WITH_NULL_SHA";
	private static final Logger logger = LoggerFactory.getLogger(TCPServer.class); 
	private TCPBCConfig config;

	private InetSocketAddress socketAddress;
	private TCPServerProtocolHandler serverHandler;
	private IoAcceptorConfig acceptorConfig;
	private IoAcceptor acceptor = null;
	private HashMap<String, IoSession> correlatedSessions;	
	
	public void init(){
		
		logger.info("-- TCP Server Component Init --");
		
		TCPBCConfig config = new TCPBCConfig();
		config.setConnNumber(propertyConfigurator.getInteger("connection.number"));
		config.setLocalPort(propertyConfigurator.getInteger("local.port"));
		config.setLocalAddress(propertyConfigurator.getString("local.address"));

//		config.setPointType(propertyConfigurator.getString("point.type"));
//		config.setPointMode(propertyConfigurator.getString("point.mode"));
		config.setOutNmEnvelope(propertyConfigurator.getString("out.envelope"));
		config.setInNmEnvelope(propertyConfigurator.getString("in.envelope"));
		config.setBase64decode(propertyConfigurator.getBoolean("base64.decode"));
		config.setBase64encode(propertyConfigurator.getBoolean("base64.encode"));
		config.setDefaultMep(propertyConfigurator.getString("mep"));
		
		config.setIncomingWrapper(propertyConfigurator.getString("incoming.wrapper"));
		config.setIncomingHeader(propertyConfigurator.getString("incoming.header"));
		config.setIncomingTrailer(propertyConfigurator.getString("incoming.trailer"));
		config.setStripWrapping(propertyConfigurator.getBoolean("strip.wrapping"));
		config.setIncomingEndianness(propertyConfigurator.getString("incoming.endianness"));
	
		config.setOutgoingWrapper(propertyConfigurator.getString("outgoing.wrapper"));
		config.setOutgoingHeader(propertyConfigurator.getString("outgoing.header"));
		config.setOutgoingTrailer(propertyConfigurator.getString("outgoing.trailer"));
		config.setOutgoingEndianness(propertyConfigurator.getString("outgoing.endianness"));
		
		config.setLogConnections(propertyConfigurator.getBoolean("log.connections"));
		config.setLogData(propertyConfigurator.getBoolean("log.data"));
		config.setLogDataAsHex(propertyConfigurator.getBoolean("log.dataAsHex"));

		config.setConnectionLogFileName(propertyConfigurator.getString("log.connections.filename"));
		config.setExtraInformation(propertyConfigurator.getString("log.extrainfo"));
		
		config.setUseSSL(propertyConfigurator.getBoolean("connection.usessl"));
		
		if(config.isUseSSL()) {
			config.setUseSSLClientMode(propertyConfigurator.getBoolean("connection.clientssl"));

			config.setKeyStoreFileName(propertyConfigurator.getString("keystore.store"));
			config.setKeyStorePassword(propertyConfigurator.getString("keystore.password"));
			config.setKeyStoreType(propertyConfigurator.getString("keystore.type"));
		}
		this.config = config;
	}	
	
	/**
     * Start the MINA server; setting:
     * <ul>
     * 	<li>the 'codec' that recognise messages of the TCP protocol unwrapping received bytes</li>
     * 	<li>optionally add ssl support</li>
     * 	<li>the TCPServerProtocolHandler that manage messages identify by the 'codec'</li>
     * </ul>
     */
    public void start() throws Exception {
    	super.start();
        
        correlatedSessions = new HashMap<String, IoSession>();
        
        // Suggested in MINA FAQ
        ByteBuffer.setUseDirectBuffers(false);
        ByteBuffer.setAllocator(new SimpleByteBufferAllocator());
        
        logger.debug("start -> Byte Buffer.isUseDirectBuffer -> ["+ByteBuffer.isUseDirectBuffers()+"]");
        logger.debug("start -> Byte Buffer.Allocator Class Name -> ["+ByteBuffer.getAllocator().getClass().getName()+"]");
       

        if (config.getConnNumber() > 0) {
        	acceptor = new SocketAcceptor(config.getConnNumber(), new NewThreadExecutor());
        } else {
        	acceptor = new SocketAcceptor();
        }
        acceptorConfig = new SocketAcceptorConfig();
        
        ((SocketAcceptorConfig)acceptorConfig).setReuseAddress(true);
        ((SocketAcceptorConfig)acceptorConfig).setDisconnectOnUnbind(true);
        ((SocketAcceptorConfig)acceptorConfig).setBacklog(config.getListenBacklog());
        
        
        // This is the filters pipeline
        DefaultIoFilterChainBuilder chain = acceptorConfig.getFilterChain();
        
        if (logger.isDebugEnabled()) {
        	addLogger( chain );
        }
        
        if (config.isUseSSL()) {
        	addSSLSupport(chain);
        }

        chain.addLast( "codec", new ProtocolCodecFilter(
            	new TCPMsgCodecFactory(config)));
        
        socketAddress = new InetSocketAddress(
            	InetAddress.getByName(this.config.getLocalAddress()),
            	this.config.getLocalPort() );
        serverHandler = new TCPServerProtocolHandler(this);
        
        // Bind
        startAccept();

        logger.info( "Listening on port " + this.config.getLocalPort() );
    }
    
    public void stop() throws Exception {
    	super.stop();
        if (serverHandler != null) {
        	serverHandler.dispose();
        }
        acceptor.unbindAll();
    }
    
    public void pauseAccept() {
    	acceptor.unbindAll();
    }
    
    public void startAccept() throws IOException {
    	logger.debug("Acceptor managing socket? "+acceptor.isManaged(socketAddress));
    	if (!acceptor.isManaged(socketAddress)) {
   			acceptor.bind(socketAddress, serverHandler, acceptorConfig);
    	}
    	logger.debug("Acceptor managing socket? "+acceptor.isManaged(socketAddress));
    }

    public void process(Exchange exchange)  {
    	// As we act as a consumer (we just send JBI exchanges)
    	// we will receive responses or DONE / ERROR status here
    	logger.debug("Received ME: ", exchange);


    	if (exchange.getStatus() == Status.Done) {
    		// This should never happen
    		logger.debug("Received an unattended DONE message");
    		return;
    	}

    	// Better restrict sync area to the parts that really need it
    	synchronized (correlatedSessions) {

    		if (config.getDefaultMep().equals(
    				SpagicConstants.IN_OUT_MEP)) {
    			String correlationId=null;

    			try {

    				correlationId = (String)exchange.getProperty(SpagicConstants.CORRELATION_ID);
    				if (correlationId == null) {
    					logger.warn("CorrelationId null. The client will not receive response");
    					return;
    				}
    				IoSession session = null;
    				session = correlatedSessions.get(correlationId);

    				if (session == null) {
    					throw new IllegalStateException("CorrelationID ["+correlationId+"] does not identify any TCP session");
    				}
    				if (!session.isConnected()) {
    					logger.warn("The tcp session is not connected The client will not receive response. " +
    							"CorrelationId="+correlationId);
    					correlatedSessions.remove(correlationId);
    					return;
    				}

    				if (exchange.getStatus() == Status.Active) {

    					// Check for errors
    					if (exchange.getFault(false) != null) {
    						logger.warn("Received an error: no response is produced", exchange.getFault());
    						return;
    					}

    					Message out = exchange.getOut(false);

    					// TODO: On debug, check the correctness of the xpath query...
    					Object content = NMUtils.retrieveMessageInEnvelope(out,
    							(config.isUseOutNMEnvelope()) ? 
    									config.getOutNmEnvelope()
    									: "/");

    					logger.debug("Sending response for correlationId: "
    							+ correlationId + " msg: " + content);

    					assert session != null && session.isConnected();
    					if (config.isBase64decode()) {

    						session.write(Base64
    								.decodeBase64((byte[]) content));

    					} else {
    						session.write(content);
    					}
    				} else if (exchange.getStatus() == Status.Error) {
    					logger.warn("Received an error: no response is produced");
    					return;
    				} else {
    					logger.warn("Received an unexpected state: no response is produced");
    					return;
    				}
    			}catch (Exception e) {
    				throw new IllegalStateException(e.getMessage(),e);
    			} finally {
    				// terminate the exchange

//    				exchange.setStatus(Status.Done);
    				done(exchange);
    			}
    		}
    	}    	
    }
	
    
    
    

    private void addLogger( DefaultIoFilterChainBuilder chain ) throws Exception
    {
    	
        chain.addLast( "logger", new LoggingFilter() );
        logger.info( "Logging ON" );
    }
    
    private void addSSLSupport( DefaultIoFilterChainBuilder chain )
            throws Exception
    {
    	SSLContext context = TCPBCSSLContextFactory
        			.getInstance( config );
        SSLFilter sslFilter = new SSLFilter( context );
        sslFilter.setNeedClientAuth(config.isUseSSLClientMode());
        
        // Create a SSLEngine to get the supported cipher suites
        SSLEngine engine = context.createSSLEngine();
        
        // Get enabled and supported cipher suites
        String[] supportedCipherSuites = engine.getSupportedCipherSuites();
        String[] enabledCipherSuites = engine.getEnabledCipherSuites();
        List<String> supportedCipherSuitesList = Arrays.asList(supportedCipherSuites);
        List<String> enabledCipherSuitesList = Arrays.asList(enabledCipherSuites);
        
        // Check if supported and enabled
        if (!enabledCipherSuitesList.contains(SSL_RSA_WITH_NULL_SHA)) {
        	// Enable it, if it's supported
        	if (supportedCipherSuitesList.contains(SSL_RSA_WITH_NULL_SHA)) {
        		String[] newEnabledCipherSuites = new String[enabledCipherSuitesList.size() + 1];
        		enabledCipherSuitesList.toArray(newEnabledCipherSuites);
        		newEnabledCipherSuites[enabledCipherSuitesList.size()] = SSL_RSA_WITH_NULL_SHA;
        		
        		sslFilter.setEnabledCipherSuites(newEnabledCipherSuites);
        		logger.info("SSL_RSA_WITH_NULL_SHA enabled for TCP component");
        	} else {
        		logger.warn("SSL_RSA_WITH_NULL_SHA not enabled and not supported for TCP component");
        	}
        }
        
        chain.addLast( "sslFilter", sslFilter );
        logger.info( "SSL ON" );
    }    

	public TCPBCConfig getConfig() {
		return config;
	}
	
	public synchronized void addCorrelatedSession(String id, IoSession session) {
		logger.debug("Adding correlationID: "+id);
		correlatedSessions.put(id, session);
	}
	
	public synchronized void removeCorrelatedSession(String id) {
		logger.debug("Removing correlationID: "+id);
		correlatedSessions.remove(id);
	}

	public Exchange createExchange() {
		String mep = config.getDefaultMep();
		Exchange rv=null;
		
		if(CommonConfig.DEFINOUTMEP.equals(mep)) {
			rv = createInOutExchange();
		}
		else if (CommonConfig.DEFINMEP.equals(mep)) {
			rv = createInOnlyExchange();
		}
		else
			throw new IllegalStateException("MEP should be either in-only or in-out");
		

		return rv;
	}


}
