package org.spagic3.core;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;
import org.spagic3.nmr.api.Status;
import org.spagic3.nmr.core.MessageImpl;

/**
 * Class <code>AbstractSpagicConnector</code> Abstract implementation
 * of the IConnector interface. Provides the basic implementation of
 * the start() and stop() method that <b>MUST</b> be invoked by
 * subclasses in order to keep the monitoring working properly.
 *
 * @author <a href="mailto:saint@eng.it">Gian Uberto Lauri</a>
 * @version $Revision$
 */
public abstract class AbstractSpagicConnector extends AbstractSpagicService implements IConnector {
	private static Logger logger = LoggerFactory.getLogger(AbstractSpagicConnector.class);
		
	/**
	 * <code>process</code> method invoked when the connector receives
	 * an exchange.
	 *
	 * @param exchange an <code>Exchange</code> the exchange received
	 */
	public void process(Exchange exchange){
		// Skip done exchanges
		if (exchange.getStatus() == Status.Done) {
			return;
			// Handle error exchanges
		} else if (exchange.getStatus() == Status.Error) {
			return;
		}
      
      try {
          Exchange outExchange = null;
          Message in = exchange.getIn();
          Message out = null;
          if (ExchangeUtils.isInAndOut(exchange)) {
              out = exchange.getOut(true);
          } else {
        	  // InOnly: build an output exchange if this connector has a target configured
        	  if (this.target != null) {
                  outExchange = createInOnlyExchange();
                  outExchange.setProperty(SpagicConstants.SPAGIC_SENDER, getSpagicId());
                  String processCorrelationId = (String)exchange.getProperty(SpagicConstants.CORRELATION_ID);
                  if (processCorrelationId != null) {
                      outExchange.setProperty(SpagicConstants.CORRELATION_ID, processCorrelationId);
                  }
                  out = exchange.getIn(true);
        	  }
          }
          
          if (out != null) {
        	  copyMessage(in, out);
              copyPropertiesAndAttachments(exchange, in, out);
          }
          
          if (run(exchange, in, out)) {
				if (ExchangeUtils.isInAndOut(exchange)) {
					configureForResponse(exchange);
					exchange.setOut(out);
					send(exchange);
				} else {
					// *******************************************************
					// 1 - Send the Out Exchange
					// *******************************************************
					if (outExchange != null) {
						outExchange.setIn(out);
						send(outExchange);
					}

					// *******************************************************
					// 2 - Send Back the original Exchange with status DONE
					// *******************************************************
					done(exchange);
				}
          } else {
              done(exchange);
          }
      } catch (Exception e) {
    	  logger.error("Error in connection execution", e);
	      fail(exchange, e);
      }
	}
	
	/**
	 * <code>run</code> alternative method to process an exchange,
	 * with messages already extracted. Not implemented.
	 *
	 * @param exchange an <code>Exchange</code> the exchange received
	 * @param in a <code>Message</code> the in message
	 * @param out a <code>Message</code> the out message
	 * @return a <code>boolean</code> never returned
	 * @exception Exception always thrown as UnsupportedOperationException
	 */
	public boolean run(Exchange exchange, Message in, Message out) throws Exception {
		 throw new UnsupportedOperationException("run method not implemented on connector");
	 }

	 /**
	  * Copy all properties and attachments from one message to another
	 * @param exchange Original exchange
	 * @param in Input message
	 * @param out Output message
	 * @throws Exception
	 */
	protected void copyPropertiesAndAttachments(Exchange exchange, Message in,
			Message out) throws Exception {
		out.setHeaders(in.getHeaders());

		String attachmentsIn = ((MessageImpl) in)._encodeAttachments();
		((MessageImpl) out)._decodeAttachments(attachmentsIn);
	}
	
	/**
	 * Copy the content of the in message to the out message.
	 * @param in Input message
	 * @param out Output message
	 * @throws Exception
	 */
	protected void copyMessage(Message in, Message out) throws Exception {
		out.setBody(in.getBody());
	}

}
