package org.spagic3.core;

import java.util.Dictionary;
import java.util.concurrent.atomic.AtomicReference;

import org.osgi.service.component.ComponentContext;
import org.osgi.service.event.Event;
import org.osgi.service.event.EventHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.core.routing.IMessageRouter;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Message;
import org.spagic3.nmr.api.Pattern;
import org.spagic3.nmr.api.Status;
import org.spagic3.util.properties.PropertyConfigurator;

public  abstract class AbstractSpagicService implements ISpagicService, EventHandler {

	private static Logger logger = LoggerFactory.getLogger(AbstractSpagicService.class);
	
	protected String spagicId = null;
	protected String target = null;
	protected PropertyConfigurator propertyConfigurator = null;
	private boolean started;
	private final AtomicReference<IMessageRouter> router = new AtomicReference<IMessageRouter>(null); 
	// Saint - Thu Mar  4 11:57:04 2010
	// 
	// Added local reference to the component context
	protected ComponentContext context = null;
	//  </Saint>
	
	/**
	 * <code>activate</code> Activation of the connector. Calls the
	 * start() method which, in turn, must call super.start().
	 *
	 * @param componentContext a <code>ComponentContext</code> the
	 * context of the component.
	 */
	@SuppressWarnings("unchecked")
	protected void activate(ComponentContext componentContext){
		// Saint - Thu Mar  4 11:57:23 2010
		// 
		// Added local reference to the component context
		context = componentContext;
		//  </Saint>
		Dictionary propertiesDictionary = componentContext.getProperties();
		this.propertyConfigurator = new PropertyConfigurator(propertiesDictionary);
		
		this.spagicId = (String)this.propertyConfigurator.getString("spagic.id");
		this.target = (String)this.propertyConfigurator.getString("target",null);
		
		init();
		
		try{
			start();
		}catch (Exception e) {
			logger.error(e.getMessage(), e);
		}
	}
	
	/**
	 * <code>deactivate</code> Deactivation of the service. Calls
	 * the stop() method which, in turn, must call super.stop().
	 *
	 * @param componentContext a <code>ComponentContext</code> the
	 * context of the component.
	 */
	protected void deactivate(ComponentContext componentContext){
		// Saint - Thu Mar  4 11:57:56 2010
		// 
		// Added local reference to the component context
		context = null;
		//  </Saint>
		beforeDestroy();

		try{
			stop();	
		}catch (Exception e) {
			logger.error(e.getMessage(), e);
		}
	}
	
	public void beforeDestroy(){
		
	}
	
	public void init(){
		
	}
	
	
	public IMessageRouter getMessageRouter() {
		return this.router.get();
	}

	
	public void unsetMessageRouter(IMessageRouter router) {
		this.router.compareAndSet(router, null);
	}
	
	
	public String getSpagicId() {
		return this.spagicId;
	}

	

	
	public void setMessageRouter(IMessageRouter router) {
		this.router.set(router);
		
	}

	public boolean isSpagicInternalEvent(Event event){
	
		return ( event.getProperty(SpagicConstants._IS_INTERNAL_EVENT) != null 
					? (Boolean)event.getProperty(SpagicConstants._IS_INTERNAL_EVENT) 
							: false );
	}
	
	public void handleInternalEvent(Event event){
		// Do Nothing
		// Must be handled by components
	}
	
	public void handleEvent(Event event) {
		logger.info("Service["+getSpagicId()+"] -> Received Event " + event);
		if (isSpagicInternalEvent(event)){
			logger.info("****** INTERNAL EVENT -> Service["+getSpagicId()+"] -> Received Event " + event);
			handleInternalEvent(event);
			return;
		}
		//logger.info("Service["+getSpagicId()+"] -> Received Event " + event);
		Exchange exchange = ExchangeUtils.fromEvent(event);
		logger.info("Service["+getSpagicId()+"] -> Received Exchange ["+exchange.getId()+"] ["+exchange.getStatus()+"]" );
		try{
			process(exchange);
		}catch (Throwable e) {
			logger.error("Exception Processing Exchange ");
			logger.error(e.getMessage(), e);
		}
	}
	
	
	
	public void send(Exchange exchange){
		if (!ExchangeUtils.isSync(exchange)){
			asyncSend(exchange);
		}
		
	}
	
	protected void asyncSend(Exchange exchange){
		try{
			getMessageRouter().send(exchange);
		}catch (Throwable e) {
			logger.error("Exception Sending Exchange ");
			throw new IllegalStateException(e);
		}
	}
	
	public void fail(Exchange exchange, Exception error) {
			configureForResponse(exchange);
			if (ExchangeUtils.isInOnly(exchange)) {
	            exchange.setError(error);
	        } else {
	            Message fault = exchange.getFault(true);
	            fault.setBody("<error>"+error.getMessage()+"</error>");
	            exchange.setFault(fault);
	        }
	        send(exchange);
	 }
	
	public void done(Exchange exchange) {
		configureForResponse(exchange);
		exchange.setStatus(Status.Done);
		send(exchange);
	}
	 
	 
	
	public void configureForResponse(Exchange exchange){
		String originalSender = (String)exchange.getProperty(SpagicConstants.SPAGIC_SENDER);
		exchange.setProperty(SpagicConstants.SPAGIC_SENDER, getSpagicId());
		exchange.setProperty(SpagicConstants.SPAGIC_TARGET, originalSender);
	}
	
	/**
	 * This method is useful if the exchange is not created by the component
	 * itself, but by an external class such as a marshaller that is
	 * unaware of the component target.
	 * @param exchange The exchange to fill with sender and target information
	 */
	public void configureExchangeTarget(Exchange exchange) {
		exchange.setProperty(SpagicConstants.SPAGIC_SENDER, getSpagicId());
		exchange.setProperty(SpagicConstants.SPAGIC_TARGET, this.target);
	}
	
	protected Exchange createInOnlyExchange(){
		return ExchangeUtils.createExchange(getSpagicId(), this.target, Pattern.InOnly);
	}
	
	protected Exchange createInOutExchange(){
		return ExchangeUtils.createExchange(getSpagicId(), this.target, Pattern.InOut);
	}
	
	public boolean isStarted() throws Exception {
		return started;
	}

	public boolean isStopped() throws Exception {
		return ! started;
	}

	public void start() throws Exception {
		started =  true;
		
	}

	public void stop() throws Exception {
		started = false;
	}
	
	

}
