package org.spagic3.core.registry;

import java.util.ArrayList;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicReference;

import org.jgroups.Address;
import org.jgroups.JChannel;
import org.jgroups.ReceiverAdapter;
import org.jgroups.View;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.service.component.ComponentContext;
import org.osgi.service.event.Event;
import org.osgi.service.event.EventAdmin;
import org.osgi.service.event.EventHandler;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.core.SpagicNode;

public class SpagicRegistry extends ReceiverAdapter implements ISpagicRegistry, EventHandler, ServiceListener {

	public static String REGISTRY_EVENT_TYPE = "REGISTRY_EVENT_TYPE";
	public static String REGISTRY_EVENT_ADDED = "REGISTRY_EVENT_ADD";
	public static String REGISTRY_EVENT_REMOVED = "REGISTRY_EVENT_REMOVED";
	public static String REGISTRY_EVENT_SPAGIC_ID  = "REGISTRY_EVENT_SPAGIC_ID";
	public static String REGISTRY_EVENT_FROM_NODE = "REGISTRY_EVENT_FROM_NODE";
	public static String REGISTRY_EVENT_NODE_UP = "REGISTRY_EVENT_NODE_UP";
	
	private CopyOnWriteArrayList<String> localServices = new CopyOnWriteArrayList<String>();
	
	
	private JChannel channel = null;
	
	// Node Identifier, Services for Node
	private ConcurrentHashMap<String, List<String>> nodes = new ConcurrentHashMap<String, List<String>>();

	private AtomicReference<EventAdmin> eventAdmin = new AtomicReference<EventAdmin>();
	
	protected void activate(ComponentContext componentContext){
		componentContext.getBundleContext().addServiceListener(this);
	}
	
	protected void deactivate(ComponentContext componentContext){
		componentContext.getBundleContext().removeServiceListener(this);
		
	}
	
	/*
	protected boolean isUniversalID(String spagicId){
		return ( spagicId.indexOf("@") > 0);
	}
	
	protected String getRelativeId(String spagicUniversalID){
		String[] part =  spagicUniversalID.split("@");
		return part[0];
	}
	
	
	protected String getNode(String spagicUniversalID){
		String[] part =  spagicUniversalID.split("@");
		return part[1];
	}
	*/
	
	
	
	protected boolean isRegisteredRemotedOnNode(String spagicId, String node){
		List<String> servicesOnNode = nodes.get(node);
		return servicesOnNode != null && (!servicesOnNode.isEmpty()) && (servicesOnNode.contains(spagicId));
	}
	
	
	public boolean isRegistered(String spagicId) {
		
			return isRelativeIdRegistered(spagicId);
	}
	
	/* no more need
	public boolean isUniversaIdRegistered(String spagicId) {
			String relativeId = getRelativeId(spagicId);
			String node = getNode(spagicId);
			if (node.equals(currentNode())){
				
			}else{
				return isRegisteredRemotedOnNode(relativeId,node);
			}
	}*/
		
	public boolean isRelativeIdRegistered(String relativeId){
		boolean localReg = localServices.contains(relativeId);
		if (!localReg){
			for (String n : nodes.keySet()){
				if (isRegisteredRemotedOnNode(relativeId, n))
					return true;
			}
			return false;
		}else{
			return localReg;
		}
	}

	
	public void bind(EventAdmin ea){
		eventAdmin.set(ea);
		if (SpagicNode.isClusteredEnabled()){
			try{
				channel = new JChannel();
				channel.setName(SpagicNode.id());
				channel.setReceiver(this);
				channel.connect("SpagicCluster");
				//channel.getState(null, 10000);
			}catch (Exception e) {
				e.printStackTrace();
			}
		}
		nodeUp();
		snapshot();
		
	}

	public void unbind(EventAdmin ea){
		if (SpagicNode.isClusteredEnabled()){
			channel.close();
		}
		eventAdmin.compareAndSet(ea,null);
		
	}
	
	public void serviceChanged(ServiceEvent event) {
		String spagicId = getSpagicId(event);
		if (spagicId != null){
			  if (event.getType() == ServiceEvent.REGISTERED) {
		            localServices.add(spagicId);
		            notify(spagicId, REGISTRY_EVENT_ADDED);
		      }else if (event.getType() == ServiceEvent.UNREGISTERING){
		            localServices.remove(spagicId);
		            notify(spagicId, REGISTRY_EVENT_REMOVED);
		      } else if (event.getType() == ServiceEvent.MODIFIED){
		           
		      }
		}	
	}
	
	private String getSpagicId(ServiceEvent event){
		return (String)event.getServiceReference().getProperty(SpagicConstants.SPAGIC_ID_PROPERTY);
	}
	
	public void snapshot() {
		for (String localString : localServices){
			notify(localString, REGISTRY_EVENT_ADDED);
		}
	}
	
	public void notify(String spagicId, String evType) {
		EventAdmin ea = eventAdmin.get();
		if (ea != null){
			Dictionary<String, String> evProperties = new Hashtable<String, String>();
			evProperties.put(REGISTRY_EVENT_TYPE, evType);
			evProperties.put(REGISTRY_EVENT_FROM_NODE, System.getProperty(SpagicConstants.SPAGIC_NODE_IDENTIFIER));
			evProperties.put(REGISTRY_EVENT_SPAGIC_ID, spagicId);
			Event ev = new Event("SPAGIC/REGISTRY", evProperties );
			ea.postEvent(ev);
		}
		
	}
	
	public void nodeUp() {
		EventAdmin ea = eventAdmin.get();
		if (ea != null){
			Dictionary<String, String> evProperties = new Hashtable<String, String>();
			evProperties.put(REGISTRY_EVENT_TYPE, REGISTRY_EVENT_NODE_UP);
			evProperties.put(REGISTRY_EVENT_FROM_NODE, System.getProperty(SpagicConstants.SPAGIC_NODE_IDENTIFIER));
			Event ev = new Event("SPAGIC/REGISTRY", evProperties );
			ea.postEvent(ev);
		}
		
	}
	
	public boolean isFromMySelf(Event event){
		String from = (String)event.getProperty(REGISTRY_EVENT_FROM_NODE);
		return SpagicNode.id().equals(from);
	}
	
	
	public void registeredServiceFromNode(String spagicId, String from){
		if (!(nodes.containsKey(from))){
			List<String> servicesOnNode = new ArrayList<String>();
			servicesOnNode.add(spagicId);
			nodes.put(from,servicesOnNode);
		}else{
			nodes.get(from).add(spagicId);
		}
	}
	
	public void unregistered(String spagicId, String from){
		if (nodes.containsKey(from)){
			nodes.get(from).remove(spagicId);
			if (nodes.get(from).size() == 0){
				nodes.remove(from);
			}
		}
	}
	// Events Received by Event Admin -- Other Nodes Registration
	public void handleEvent(Event event) {
		
		String evType = (String)event.getProperty(REGISTRY_EVENT_TYPE);
		String from = (String)event.getProperty(REGISTRY_EVENT_FROM_NODE);
		System.out.println("SPAGIC_REGISTRY ->  REGISTRY EVENT FROM ["+from+"] -> ["+evType+"]");
		if (!isFromMySelf(event)){
			
			String spagicId  = (String)event.getProperty(REGISTRY_EVENT_SPAGIC_ID);
			if (evType.equalsIgnoreCase(REGISTRY_EVENT_ADDED)){
				registeredServiceFromNode(spagicId, from);
			} else if (evType.equalsIgnoreCase(REGISTRY_EVENT_REMOVED)){
				unregistered(spagicId, from);
			}  else if (evType.equalsIgnoreCase(REGISTRY_EVENT_NODE_UP)){
				snapshot();	
			}
		}else{
			System.out.println("SPAGIC_REGISTRY ->  FROM MYSELF IGNORE IT :-)");
		}
	}

	public String getServiceNode(String spagicId) {
			if (localServices.contains(spagicId)){
				return SpagicNode.id();
			}else{
				for ( String node : nodes.keySet()){
					if (isRegisteredRemotedOnNode(spagicId, node))
						return node;	
				}
				return null;
			}
	}

	public List<String> getSnaphshot() {
		List<String> snapshot = new ArrayList<String>();
		
		
		for ( String local : localServices){
			snapshot.add(local + "@" + System.getProperty(SpagicConstants.SPAGIC_NODE_IDENTIFIER) );
		}
		
		for ( String key : nodes.keySet()){
			List<String> services = nodes.get(key);
			for ( String s : services){
				snapshot.add(s + "@" + key);
			}
		}
		
		return snapshot;
	}

	public boolean isRegisteredLocally(String spagicId){
	
			return localServices.contains(spagicId);
		
	}

	 public void viewAccepted(View newView) {
		 
	        System.out.println("** view: " + newView);
	        Vector<Address> viewMembers = newView.getMembers();
	        List<String> viewMembersAsString = new ArrayList<String>();
	        
	        boolean newNodeAdded = false;
	        for (Address address : viewMembers){
	        	String newViewNode = address.toString();
	        	viewMembersAsString.add(newViewNode);
	        	if (newViewNode.equals(SpagicNode.id()))
	        		continue;
	        	if (!nodes.contains(newViewNode)){
	        		nodes.put(newViewNode, new ArrayList<String>());
	        		newNodeAdded = true;
	        	}
	        }
	        if (newNodeAdded)
	        	snapshot();
	        
	        List<String> toRemove = new ArrayList<String>();
	        for (String node : nodes.keySet()){
	        	if (!viewMembersAsString.contains(node)){
	        		toRemove.add(node);
	        	}
	        }
	        
	        for (String n : toRemove){
	        	nodes.remove(n);
	        }
	 }
}
