package org.spagic3.core.routing;

import java.util.UUID;
import java.util.concurrent.atomic.AtomicReference;

import org.osgi.framework.ServiceReference;
import org.osgi.service.event.EventAdmin;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.core.ExchangeUtils;
import org.spagic3.core.SpagicNode;
import org.spagic3.core.registry.ISpagicRegistry;
import org.spagic3.nmr.api.Exchange;
import org.spagic3.nmr.api.Status;

public class SpagicRouter implements IMessageRouter {
	
	private static Logger logger = LoggerFactory.getLogger(SpagicRouter.class);
	
	private AtomicReference<EventAdmin> localEventAdmin = new AtomicReference<EventAdmin>();
	private AtomicReference<EventAdmin> distributedEventAdmin = new AtomicReference<EventAdmin>();
	
	private ISpagicRegistry registry = null;
	private IDynamicRouter dynamicRouter = null;
	
	
	
	
	public void send(Exchange exchange) throws Exception {
		String sender = (String)exchange.getProperties().get(SpagicConstants.SPAGIC_SENDER);
		
		// Target Configuration
		String target = (String)exchange.getProperties().get(SpagicConstants.SPAGIC_TARGET);
		
		logger.info("Spagic Router SEND EXCHANGE ["+exchange.getId()+"] ["+sender+"] ["+target+"] Status ["+exchange.getStatus().toString()+"]");

        String correlationId = (String)exchange.getProperty(SpagicConstants.CORRELATION_ID);
        if (correlationId == null) {
        	correlationId = UUID.randomUUID().toString();
        	exchange.setProperty(SpagicConstants.CORRELATION_ID, correlationId);
        }

		// If target is not configured as a properties of the exchange
		// Se if there's a connector routing map
		if (target == null){
			target = dynamicRouter.getTarget(sender);
			exchange.setProperty(SpagicConstants.SPAGIC_TARGET, target);
		}
		
		if (target == null)
			throw new Exception(" No Target Configured ");
		
		if (!registry.isRegistered(target)){
			String error = "Target Service ["+target+"] for exchange is not registered";
			sendBack(exchange, error, sender, target);
		}
		
		route(exchange, target);
	}
		
	protected void route(Exchange exchange, String target){
		// If Service is registered locally use local EventAdmin Policy
		exchange.setProperty(SpagicConstants.EXCHANGE_SENDER_NODE, System.getProperty(SpagicConstants.SPAGIC_NODE_IDENTIFIER));
		EventAdmin ea = null;
		if (registry.isRegisteredLocally(target)){
			exchange.setProperty(SpagicConstants.EXCHANGE_TARGET_NODE, SpagicNode.id());
			ea = localEventAdmin.get();
			if (ea == null){
				ea = distributedEventAdmin.get();
			}
		}else if (registry.isRegistered(target)){
			String node = registry.getServiceNode(target);
			exchange.setProperty(SpagicConstants.EXCHANGE_TARGET_NODE, node);
			ea = distributedEventAdmin.get();
			if (ea == null){
				ea = localEventAdmin.get();
			}
		}
		
		if (ea != null){
			ea.postEvent(ExchangeUtils.toEvent(exchange));
		}
	}
	

	protected void sendBack(Exchange exchange, String errorMessage, String originalSender, String originalTarget){
		if (!(exchange.getStatus() == Status.Done)){
			exchange.setProperty(SpagicConstants.SPAGIC_SENDER, SpagicConstants.SPAGIC_ROUTER);
			exchange.setProperty(SpagicConstants.SPAGIC_TARGET, originalSender);
			if (ExchangeUtils.isInAndOut(exchange)){
				exchange.getFault().setBody("<error>"+errorMessage+"</error>");
			}else{
				exchange.setError(new Exception(errorMessage));
			}
			
			route(exchange, originalSender);
			return;
		}
	}
	
	
	public void bindLocalEventAdmin(EventAdmin ea) {
		localEventAdmin.set(ea);
	}
	
	public void unbindLocalEventAdmin(EventAdmin ea) {
		localEventAdmin.compareAndSet(ea, null);
	}
	
	
	public void bindDistributedEventAdmin(EventAdmin ea) {
		distributedEventAdmin.set(ea);
	}
	
	public void unbindDistributedEventAdmin(EventAdmin ea) {
		distributedEventAdmin.compareAndSet(ea, null);
	}

	public boolean isDistributedEventAdmin(ServiceReference sr){
		String eaType = (String) sr.getProperty(SpagicConstants.EVENT_ADMIN_TYPE);
		if (eaType == null){
			return false;
		}else{
			return eaType.equalsIgnoreCase(SpagicConstants.DISTRIBUTED_EVENT_ADMIN_TYPE);
		}
	}
	

	

			
	
	
	
	
	
	public void unbindRegistry(ISpagicRegistry spagicRegistry){
		this.registry = null;
	}
	
	public void bindRegistry(ISpagicRegistry spagicRegistry){
		this.registry = spagicRegistry;
	}

	
	
	public void bindDynamicRouter(IDynamicRouter dr){
		this.dynamicRouter = dr;
	}

	public void unbindDynamicRouter(IDynamicRouter dr){
		this.dynamicRouter = null;
	}


	public boolean isServiceRegistered(String target) throws Exception {
		// TODO Auto-generated method stub
		return registry.isRegistered(target);
	}
}
