/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic3.databaseManager;

import java.util.Map;

import org.spagic.metadb.model.Component;
import org.spagic.metadb.model.Service;
import org.spagic.metadb.model.ServiceInstance;

public interface IDatabaseManager {
	
	/**
	 * Retrieves the definition of a specified Component
	 * @param componentName
	 * @return The component found, or null if not found
	 */
	public Component getComponentByName(String componentName);

	/**
	 * Retrieves the latest definition of a specified Service
	 * @param serviceId
	 * @return The service found, or null if not found
	 */
	public Service getServiceById(String serviceId);
	
	/**
	 * Retrieves the definition of a specified Service with a specific version
	 * @param serviceId
	 * @param version
	 * @return The service found, or null if not found
	 */
	public Service getServiceByIdAndVersion(String serviceId, String version);
	
	/**
	 * Retrieves a specified Service instance
	 * @param serviceInstanceId Unique id of the service instance
	 * @return The found service instance, or null if not found
	 */
	public ServiceInstance getServiceInstance(Long serviceInstanceId);
	
	/**
	 * Retrieves a specified Service instance
	 * @param exchangeID The message id
	 * @return The service instance found, or null if not found
	 */
	public ServiceInstance getServiceInstanceByMessageId(String exchangeID);
	
	/**
	 * Retrieves a specified Service instance
	 * @param serviceId The related service
	 * @param exchangeID The correlation id
	 * @return The service instance found, or null if not found
	 */
	public ServiceInstance getServiceInstanceByCorrelationId(String serviceId, String exchangeID);
	
	/**
	 * Creates a new service instance
	 * @param serviceId Service Id
	 * @param exchangeID Exchange Id
	 * @param request Request sent to the service instance (can be null)
	 * @param response Response sent by the service instance (can be null)
	 * @return The new service instance
	 */
	public ServiceInstance createServiceInstance(String serviceId, String exchangeID, String correlationId, ServiceInstance targetServiceInstance, String request, String response);

	/**
	 * Updates an existing service instance
	 * @param serviceInstance The instance to update
	 * @param response The response to update
	 */
	public void updateServiceInstance(ServiceInstance serviceInstance);

	/**
	 * Create a new service definition with a specified serviceId, associated component and
	 * a set of generic properties
	 * @param serviceId
	 * @param componentName Factory name necessary to relate a component
	 * @param properties Set of properties
	 * @return The newly created service.
	 */
	public Service registerService(String serviceId, String componentName, Map<String, String> properties);
	
	/**
	 * Create a new component definition with a specified componentId, and
	 * a set of generic properties
	 * @param componentName
	 * @param properties
	 * @return The newly created component
	 */
	public Component registerComponent(String componentName, String componentType, Map<String, String> properties);
	
	/**
	 * Links an existing service instance to an orchestration process
	 * @param exchangeId The message id identifying the service instance
	 * @param processId The process to link
	 */
	public void relateProcessToServiceInstance(String exchangeId, Long processId);

}
