/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic3.deployer;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicReference;

import org.osgi.framework.ServiceReference;
import org.osgi.service.component.ComponentConstants;
import org.osgi.service.component.ComponentContext;
import org.osgi.service.component.ComponentFactory;
import org.osgi.service.component.ComponentInstance;
import org.osgi.service.event.EventConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.core.SpagicNode;
import org.spagic3.core.SpagicUtils;
import org.spagic3.core.routing.IDynamicRouter;
import org.spagic3.startup.service.IStartedUpService;

public class DeploymentService implements IDeploymentService {

	private static Logger logger = LoggerFactory.getLogger(DeploymentService.class);
	private static final String DATASOURCE_FACTORY_IDENTIFIER ="spagic3.datasourcefactory";

	private ConcurrentHashMap<String, ServiceReference> factories = new ConcurrentHashMap<String, ServiceReference>();
	private ConcurrentHashMap<String, ComponentInstance> componentInstances = new ConcurrentHashMap<String, ComponentInstance>();
	private ConcurrentHashMap<String, HashMap<String, Hashtable>> pendingDeployments = new ConcurrentHashMap<String, HashMap<String,Hashtable>>();
	
	private ComponentContext componentContext = null;
	private AtomicReference<IDynamicRouter> dynamicRouter = new AtomicReference<IDynamicRouter>();


	protected void activate(ComponentContext componentContext) {
		logger.info(" Deployment Service - ACTIVATED");
		this.componentContext = componentContext;
	}

	protected void deactivate(ComponentContext componentContext) {
		logger.info(" Deployment Service - DEACTIVATED");
	}

	public void addComponentFactory(ServiceReference componentFactoryReference) {
		
		String componentFactoryIdentifier = (String) componentFactoryReference
				.getProperty(ComponentConstants.COMPONENT_FACTORY);
		
		logger.info(" Component Factory ["+componentFactoryIdentifier+"] -- REGISTERED");
		this.factories.put(componentFactoryIdentifier,
				componentFactoryReference);
		
		logger.info(" Revamp Pending Deployment for ["+componentFactoryIdentifier+"]");
		HashMap<String, Hashtable> pendingDeploymentForFactory = pendingDeployments.remove(componentFactoryIdentifier);
		Hashtable pendingProp = null;
		if (pendingDeploymentForFactory != null){
			for ( String id : pendingDeploymentForFactory.keySet()){
				pendingProp = pendingDeploymentForFactory.get(id);
				String depType = (String)pendingProp.get(SpagicConstants.SPAGIC_TYPE);
				if (depType.equalsIgnoreCase(IDeploymentService.SPAGIC_DATASOURCE) ||
					depType.equalsIgnoreCase(IDeploymentService.SPAGIC_CONNECTIONFACTORY))
					internalDeploy(id, depType, componentFactoryIdentifier, pendingProp, false);
				else
					internalDeploy(id, depType, componentFactoryIdentifier, pendingProp, true);
			}
		}
	}

	public void removeComponentFactory(
			ServiceReference componentFactoryReference) {
		String componentFactoryIdentifier = (String) componentFactoryReference
				.getProperty(ComponentConstants.COMPONENT_FACTORY);
		
		logger.info(" Component Factory ["+componentFactoryIdentifier+"] -- UNREGISTERED");
		factories.remove(componentFactoryIdentifier);
	}

	private void internalDeploy(String id, 
								String spagicType,
								String factoryName, 
								Hashtable properties, 
								boolean subscribeOnBus) {
		
		logger.info("Deploying ["+spagicType+"] With ID ["+id+"] using ["+factoryName+"]");
				
		ServiceReference sr = factories.get(factoryName);
		properties.put(SpagicConstants.SPAGIC_TYPE, spagicType);
		
		if (sr == null) {
			logger.warn("Cannot Find A Component Factory [" + factoryName + "] for ["+spagicType+"] With ID ["+id+"]");
			if (pendingDeployments.get(factoryName) == null)
				pendingDeployments.put(factoryName, new HashMap<String, Hashtable>());
			pendingDeployments.get(factoryName).put(id, properties);
		} else {
			// String universalId = id + "@" + System.getProperty(SpagicConstants.SPAGIC_NODE_IDENTIFIER);
			ComponentFactory cf = (ComponentFactory) componentContext.locateService("cf", sr);

			if (subscribeOnBus){
				logger.debug("Topic Determination for Node ["+id+"]");
				String topicToSubscribe = SpagicNode.topicForService(id);
				logger.debug("Topic To subscribe for Node ["+topicToSubscribe+"]");
				topicToSubscribe = SpagicUtils.normalizeTopic(topicToSubscribe);
				properties.put(EventConstants.EVENT_TOPIC, topicToSubscribe);
			}
			
			//properties.put(SpagicConstants.SPAGIC_LOCAL_ID_PROPERTY, id);
			properties.put(SpagicConstants.SPAGIC_ID_PROPERTY, id);
			ComponentInstance ci = cf.newInstance(properties);
			componentInstances.put(id, ci);
			logger.info("["+spagicType+"] With ID ["+id+"] DEPLOYED ==");
			
		}
	}
	
	

	protected void internalUndeploy(String id) {
		
		ComponentInstance ci = componentInstances.get(id);
		
		if (ci != null){
			componentInstances.remove(id);
			ci.dispose();
		}
		

	}
	
	public void deployDatasource(String dataSourceId,  Hashtable properties) {
		internalDeploy(dataSourceId, IDeploymentService.SPAGIC_DATASOURCE, DeploymentService.DATASOURCE_FACTORY_IDENTIFIER, properties, false);
	}

	@Override
	public void deployConnector(String spagicId, String factoryName,
			Hashtable properties) {
		internalDeploy(spagicId, IDeploymentService.SPAGIC_CONNECTOR, factoryName, properties, true);	
	}

	@Override
	public void deployService(String spagicId, String factoryName,
			Hashtable properties) {
		internalDeploy(spagicId, IDeploymentService.SPAGIC_SERVICE, factoryName, properties, true);	
	}

	@Override
	public void undeployConnector(String spagicId) {
		internalUndeploy(spagicId);
	}

	@Override
	public void undeployDatasource(String dataSourceID) {
		internalUndeploy(dataSourceID);		
	}

	@Override
	public void undeployService(String spagicId) {
		internalUndeploy(spagicId);
	}

	@Override
	public void deployRoutes(List<String> updateRoutes, List<String> oldRoutes) {
		getDynamicRouter().updateRoutes(updateRoutes,oldRoutes);	
	}
	
	public IDynamicRouter getDynamicRouter() {
		return this.dynamicRouter.get();
	}

	
	public void bindDynamicRouter(IDynamicRouter dynRouter) {
		this.dynamicRouter.set(dynRouter);
		
	}

	public void unbindDynamicRouter(IDynamicRouter dynRouter) {
		this.dynamicRouter.compareAndSet(dynRouter, null);
	}

	@Override
	public void deployConnectionFactory(String connectionFactoryID, String factoryName,
			Hashtable properties) {
		internalDeploy(connectionFactoryID, IDeploymentService.SPAGIC_CONNECTIONFACTORY, factoryName, properties, false);
	}

	@Override
	public void undeployConnectionFactory(String connectionFactoryID) {
		internalUndeploy(connectionFactoryID);
	}
	
	public void bindStartUpService(IStartedUpService startedUpService){
		logger.info("STARTUP SERVICE IS OK -> CAN START DEPLOYMENT PHASE");	
	}
	
	public void unbindStartUpService(IStartedUpService startedUpService){
		logger.info("");
		
	}
}
