/**

    Copyright 2007, 2010 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic3.serviceregistration;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicReference;

import org.osgi.framework.ServiceReference;
import org.osgi.service.component.ComponentConstants;
import org.osgi.service.event.EventConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.spagic.metadb.model.Component;
import org.spagic.metadb.model.Service;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.databaseManager.IDatabaseManager;
import org.spagic3.service.model.IServiceModelHelper;

public class DynamicRegistrationService {
	
	protected Logger log = LoggerFactory.getLogger(DynamicRegistrationService.class);

	private AtomicReference<IDatabaseManager> databaseManager = new AtomicReference<IDatabaseManager>();
	private AtomicReference<IServiceModelHelper> serviceModelHelper = new AtomicReference<IServiceModelHelper>();
	
	private ConcurrentHashMap<String, HashMap<String, HashMap<String, String>>> pendingRegistrations = new ConcurrentHashMap<String, HashMap<String,HashMap<String, String>>>();
	
	public static final List<String> SERVICE_TRANSIENT_PROPERTIES = Arrays.asList(SpagicConstants.SPAGIC_FACTORYNAME,
			  SpagicConstants.SPAGIC_TYPE,
			  EventConstants.EVENT_TOPIC);

	/**
	 * Dinamically register a new service definition on the database.
	 * @param reference Reference to the new Spagic service
	 */
	public void addService(ServiceReference reference) {
		// Try to register the service into the database
		String serviceId = (String) reference.getProperty(SpagicConstants.SPAGIC_ID_PROPERTY);
		String factoryName = (String) reference.getProperty(SpagicConstants.SPAGIC_FACTORYNAME);
		
		log.debug("Checking service: " + serviceId);
		
		if (factoryName == null) {
			return;
		}

		HashMap<String, String> properties = new HashMap<String, String>(reference.getPropertyKeys().length);
		for (String propKey : reference.getPropertyKeys()) {
			Object propValue = reference.getProperty(propKey);
			log.debug(propKey + "=" + propValue);
			
			if ((propValue instanceof String) && (!SERVICE_TRANSIENT_PROPERTIES.contains(propKey))) {
				properties.put(propKey, (String) propValue);
			}
		}
		String componentName = getServiceModelHelper().getComponentName(factoryName);	
		log.debug("Component name: " + componentName);
		if ((componentName != null) && (componentName.length() != 0)) {
			registerServiceInDatabase(componentName, serviceId, properties);
		}
		
	}
	
	
	private void registerServiceInDatabase(String componentName, String serviceId, HashMap<String, String> properties){
			@SuppressWarnings("unused")
			Service service = getDbManager().registerService(serviceId, componentName, properties);
			if (service == null){
				log.warn("Service Not Registered because component ["+componentName+"] is not on database");
				log.warn("Put this on pending deployment for component["+componentName+"]");
				if (pendingRegistrations.get(componentName) == null)
					pendingRegistrations.put(componentName, new HashMap<String, HashMap<String, String>>());
				pendingRegistrations.get(componentName).put(serviceId, properties);
			}
	
	}

	public void removeService(ServiceReference reference) {
		// Do nothing
	}
	
	public void addComponent(ServiceReference reference) {
		String componentFactory = (String) reference.getProperty(ComponentConstants.COMPONENT_FACTORY);

		log.info("Checking component factory: " + componentFactory);
		
		if (componentFactory == null) {
			return;
		}

		Map<String, String> properties = new HashMap<String, String>(reference.getPropertyKeys().length);
		for (String propKey : reference.getPropertyKeys()) {
			Object propValue = reference.getProperty(propKey);
			log.debug("#### " + propKey + "=" + propValue);
			
			if (propValue instanceof String) {
				properties.put(propKey, (String) propValue);
			}
		}
		
		String componentName = getServiceModelHelper().getComponentName(componentFactory);
		String componentType = getServiceModelHelper().getComponentType(componentFactory);
		if ((componentName != null) && (componentName.length() != 0)) {
			@SuppressWarnings("unused")
			Component component = getDbManager().registerComponent(componentName, componentType, properties);				
		}
		
		log.info("Revamp Pending Registration for component ["+componentName+"]");
		HashMap<String, HashMap<String, String>> pendingRegistrationForComponent = pendingRegistrations.remove(componentName);
		if ( pendingRegistrationForComponent != null ){
			for ( String serviceId : pendingRegistrationForComponent.keySet()){
				registerServiceInDatabase(componentName, serviceId, pendingRegistrationForComponent.get(serviceId));
			}
		}
	}
	
	public void removeComponent(ServiceReference reference) {
		// Do nothing
	}

	public IDatabaseManager getDbManager() {
		return this.databaseManager.get();
	}
	
	public IServiceModelHelper getServiceModelHelper() {
		return this.serviceModelHelper.get();
	}
	
	public void bindDatabaseManager(IDatabaseManager dbManager) {
		this.databaseManager.set(dbManager);
	}

	public void unbindDatabaseManager(IDatabaseManager dbManager) {
		this.databaseManager.compareAndSet(dbManager, null);		
	}
	
	public void bindServiceModelHelper(IServiceModelHelper serviceModelHelper) {
		this.serviceModelHelper.set(serviceModelHelper);
	}

	public void unbindServiceModelHelper(IServiceModelHelper serviceModelHelper) {
		this.serviceModelHelper.compareAndSet(serviceModelHelper, null);		
	}
	
}
