/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* This file contains code modified from Apache code */
package org.spagic3.soap.interceptors.nmr;

import java.net.URI;
import java.util.Map;

import javax.activation.DataHandler;
import javax.xml.namespace.QName;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.spagic3.constants.SpagicConstants;
import org.spagic3.nmr.api.Pattern;
import org.spagic3.soap.api.Fault;
import org.spagic3.soap.api.Message;
import org.spagic3.soap.api.model.Binding;
import org.spagic3.soap.api.model.Operation;
import org.spagic3.soap.core.AbstractInterceptor;
import org.spagic3.soap.util.QNameUtil;
import org.w3c.dom.DocumentFragment;

/**

 */
public class NmrOutInterceptor extends AbstractInterceptor {

    private final boolean server;
    
    public NmrOutInterceptor(boolean server) {
        this.server = server;
    }
    
    public void handleMessage(Message message) {
        try{
        	org.spagic3.nmr.api.Message nm = message.getContent(org.spagic3.nmr.api.Message.class);

        Document doc = DocumentHelper.parseText((String)nm.getBody());
        message.setContent(Document.class,  doc);
        fromNMSAttachments(message, nm);
        fromNMSHeaders(message, nm);

        if (!server) {
            org.spagic3.nmr.api.Exchange me = message.getContent(org.spagic3.nmr.api.Exchange.class);
            Binding binding = message.get(Binding.class);
            Operation operation = binding.getOperation(me.getOperation());
            if (operation != null) {
                if (!areMepsEquals(me.getPattern(), operation.getMep())) {
                    throw new Fault("Received incorrect exchange mep.  Received " + me.getPattern()
                                    + " but expected " + operation.getMep() + " for operation "
                                    + operation.getName());
                }
                message.put(Operation.class, operation);
                message.put(org.spagic3.soap.api.model.Message.class,
                            server ? operation.getOutput() : operation.getInput());
            }
        }
        }catch (Exception e) {
			throw new RuntimeException(e);
		}
    }

    
    private URI getURIMep(org.spagic3.nmr.api.Pattern mepPattern){
    	if (mepPattern == org.spagic3.nmr.api.Pattern.InOnly)
    		return SpagicConstants.IN_ONLY_URI;
    	else if (mepPattern == org.spagic3.nmr.api.Pattern.InOut)
    		return SpagicConstants.IN_OUT_URI;
    	else
    		return null;
    }
    private boolean areMepsEquals(Pattern mepPattern, URI mep2) {
        
    	URI mep1 = getURIMep(mepPattern);
    	if (mep1 == null)
    		return false;
    	
    	String s1 = mep1 != null ? mep1.toString() : "";
        String s2 = mep2 != null ? mep2.toString() : "";
        int i1 = s1.lastIndexOf('/');
        int i2 = s2.lastIndexOf('/');
        if (i1 >= 0 && i2 >= 0) {
            return s1.substring(i1).equals(s2.substring(i2));
        }
        return false;
    }

    /**
     * Copy NormalizedMessage attachments to SoapMessage attachments
     */
    private void fromNMSAttachments(Message message, org.spagic3.nmr.api.Message normalizedMessage) {
        
    	Map<String, Object> attachmentMap = normalizedMessage.getAttachments();
        
        for ( String attachmentId : attachmentMap.keySet()){
        	 DataHandler handler = (DataHandler)attachmentMap.get(attachmentId);
             message.getAttachments().put(attachmentId, handler);
        }
        
    }

    /**
     * Copy NormalizedMessage headers to SoapMessage headers
     */
    @SuppressWarnings("unchecked")
    private void fromNMSHeaders(Message message, org.spagic3.nmr.api.Message normalizedMessage) {
        if (normalizedMessage.getHeader(SpagicConstants.PROTOCOL_HEADERS) != null) {
            Map<String, ?> headers = (Map<String, ?>) normalizedMessage.getHeader(SpagicConstants.PROTOCOL_HEADERS);
            for (Map.Entry<String, ?> entry : headers.entrySet()) {
                QName name = QNameUtil.parse(entry.getKey());
                if (name != null) {
                    message.getSoapHeaders().put(name, (DocumentFragment) entry.getValue());
                } else {
                    message.getTransportHeaders().put(entry.getKey(), (String) entry.getValue());
                }
            }
        }
    }

}
